/*
 * Copyright (c) 2000 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.app.extension.dm.dictcopy;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.CDATASection;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class DictionaryConverter
{

	private static final String DICTIONARY = "dictionary";

	private static final String VALUE = "value";

	private static final String GROUP = "group";

	private static final String BASE_VALUE = "baseValue";


	public static Map<String, List<String>> convertToRawDictionary(
			Document document, DictionaryLoaderListener listener)
	{
		NodeList nl = document.getElementsByTagName(GROUP);

		Map<String, List<String>> rawDictionary = new HashMap<String, List<String>>();

		int counter = 0;
		
		for (int i = 0; i < nl.getLength(); i++) {
			Node group = nl.item(i);

			fillWithSynonyms(rawDictionary, group);
			
			counter++;
			listener.inProgress(counter, nl.getLength());
		}

		return rawDictionary;
	}


	private static void fillWithSynonyms(Map<String, List<String>> langMap,
			Node group)
	{
		NodeList nl = group.getChildNodes();

		String baseValue = null;
		List<String> values = new ArrayList<String>();
		for (int i = 0; i < nl.getLength(); i++) {
			Node node = nl.item(i);
			if (BASE_VALUE.equals(node.getNodeName())) {				
				baseValue = getNodeText(node);
			}
			else if (VALUE.equals(node.getNodeName())) {
				values.add(getNodeText(node));
			}
		}
		
		langMap.put(baseValue, values);
	}


	private static String getNodeText(Node node)
	{
		NodeList nl = node.getChildNodes();
		
		for (int i = 0; i < nl.getLength(); i++) {
			Node nodeItem = nl.item(i);
			if (nodeItem instanceof CDATASection) {
				return ((CDATASection)nodeItem).getTextContent();
			}
		}
		
		return node.getTextContent();
	}


	public static Document convertToDOMDictionary(
			Map<String, List<String>> rawDictionary)
		throws ParserConfigurationException
	{
		Document doc = DocumentBuilderFactory.newInstance()
				.newDocumentBuilder().newDocument();

		Element dictionary = doc.createElement(DICTIONARY);

		fillWithSynonyms(doc, dictionary, rawDictionary);

		doc.appendChild(dictionary);

		return doc;
	}


	private static void fillWithSynonyms(Document doc, Element dictionary,
			Map<String, List<String>> value)
	{

		for (Entry<String, List<String>> e : value.entrySet()) {
			Element group = doc.createElement(GROUP);

			Element baseValue = doc.createElement(BASE_VALUE);
			CDATASection textCDATA = doc.createCDATASection(e.getKey());
			baseValue.appendChild(textCDATA);

			group.appendChild(baseValue);

			for (String v : e.getValue()) {
				Element valueNode = doc.createElement(VALUE);
				CDATASection valueCDATA = doc.createCDATASection(v);
				valueNode.appendChild(valueCDATA);

				group.appendChild(valueNode);
			}

			dictionary.appendChild(group);
		}
	}
}
