package pl.psnc.dlibra.app.extension.dm.dictcopy;

import java.awt.Graphics2D;
import java.awt.Window;
import java.io.File;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;

import javax.swing.JFileChooser;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;

import pl.psnc.dlibra.app.common.TextWrapper;
import pl.psnc.dlibra.app.common.component.JSystemFileChooser;
import pl.psnc.dlibra.app.common.component.extension.AbstractConfigurableExtension;
import pl.psnc.dlibra.app.common.component.inprogress.OperationInProgressManager;
import pl.psnc.dlibra.app.extension.dictionarymanager.DictionaryManager;
import pl.psnc.dlibra.app.extension.dictionarymanager.DictionaryManagerException;

/**
 */
public class DictionaryCopyManager
	extends AbstractConfigurableExtension
	implements DictionaryManager
{

	public static final String BASE_VALUE_KEY = "baseValue";

	public static final String ALTERNATIVE_VALUES_KEY = "alternativeValues";

	private JSystemFileChooser sfc;


	public DictionaryCopyManager()
	{
		DEFAULT_BUNDLE_PATH = "pl/psnc/dlibra/app/extension/dm/dictcopy/res/GUI";
		reloadResourceBundle(DEFAULT_BUNDLE_PATH);

		sfc = new JSystemFileChooser();
	}


	@Override
	public boolean isConfigurable()
	{
		return false;
	}


	/**
	 * Reloads resources using given path to bundles.
	 * @param bundlePath
	 */
	@Override
	protected void reloadResourceBundle(String bundlePath)
	{
		textsBundle = ResourceBundle.getBundle(bundlePath, Locale.getDefault());
	}


	@Override
	public String toString()
	{
		if (!Locale.getDefault().equals(textsBundle.getLocale())) {
			reloadResourceBundle(DEFAULT_BUNDLE_PATH);
		}
		return textsBundle.getString("DictionaryCopyManager.name");
	}


	public Map<String, List<String>> loadDictionary(
			Window parentWindow)
		throws DictionaryManagerException
	{
		reloadResourceBundle(DEFAULT_BUNDLE_PATH);
		Map<String, List<String>> dictionary = null;

		int result = sfc.showOpenDialog(parentWindow);
		if (result == JFileChooser.APPROVE_OPTION) {
			OperationInProgressManager manager = new OperationInProgressManager(
					parentWindow);
			DictionaryLoaderListener listener = new DictLoaderListener(manager,
					textsBundle.getString("DictionaryCopyManager.importing"));

			manager.start();
			try {
				File selFile = sfc.getSelectedFile();

				DocumentBuilder db = DocumentBuilderFactory.newInstance()
						.newDocumentBuilder();

				Document document = db.parse(selFile);

				dictionary = DictionaryConverter
						.convertToRawDictionary(document, listener);
			}
			catch (Exception e) {
				throw new DictionaryManagerException(
						new TextWrapper().wrapText(
							textsBundle.getString(
								"DictionaryCopyManager.loadErrorMessage")
									.replace("{1}", e.getMessage()),
							parentWindow.getWidth() / 2,
							(Graphics2D) parentWindow.getGraphics()));
			}
			finally {
				manager.stop();
			}
		}
		else {
			return null;
		}

		return dictionary;
	}


	public void storeDictionary(Window parentWindow,
			Map<String, List<String>> dictionary)
		throws DictionaryManagerException
	{
		reloadResourceBundle(DEFAULT_BUNDLE_PATH);
		int result = sfc.showSaveDialog(parentWindow);

		if (result == JFileChooser.APPROVE_OPTION) {
			OperationInProgressManager manager = new OperationInProgressManager(
					parentWindow);

			manager.start();
			try {
				File selFile = sfc.getSelectedFile();

				Document document = DictionaryConverter
						.convertToDOMDictionary(dictionary);

				Transformer transformer = TransformerFactory.newInstance()
						.newTransformer();
				transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION,
					"yes");
				transformer.setOutputProperty(OutputKeys.INDENT, "yes");
				transformer.transform(new DOMSource(document),
					new StreamResult(selFile));
			}
			catch (Exception e) {
				throw new DictionaryManagerException(
						new TextWrapper().wrapText(
							textsBundle.getString(
								"DictionaryCopyManager.storeErrorMessage")
									.replace("{1}", e.getMessage()),
							parentWindow.getWidth() / 2,
							(Graphics2D) parentWindow.getGraphics()));
			}
			finally {
				manager.stop();
			}
		}
		else {
			throw new DictionaryManagerException(null);
		}
	}


	public boolean supports(DictionaryAction action)
	{
		return DictionaryAction.LOAD.equals(action)
				|| DictionaryAction.STORE.equals(action);
	}
}
