package pl.psnc.dlibra.app.extension.dsurl;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;

import javax.swing.JComponent;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.log4j.Logger;

import pl.psnc.dlibra.app.common.component.extension.AbstractConfigurableExtension;
import pl.psnc.dlibra.app.extension.datasource.DataSource;
import pl.psnc.dlibra.app.extension.datasource.FileInfo;
import pl.psnc.dlibra.app.extension.datasource.FileSelectionListener;
import pl.psnc.dlibra.app.extension.datasource.LocalResourceUnavailableException;
import pl.psnc.dlibra.app.extension.datasource.SimpleFileInfo;
import pl.psnc.dlibra.app.extension.dsurl.URLInputPanel.ChangesListener;

/**
 * Data Source extension that creates an xml file with a link publication's description
 * created from URLs provided by users.
 */
public class URLDataSource
	extends AbstractConfigurableExtension
	implements DataSource, ChangesListener
{
	public final static String TEXT_BUNDLE_PATH = "pl/psnc/dlibra/app/extension/dsurl/res/GUI";

	private final static Logger logger = Logger.getLogger(URLDataSource.class);

	private File tempDir;

	private File tempFile;

	private URLInputPanel panel;

	private FileSelectionListener fileSelectionListener;


	@Override
	public void initialize(Map<String, Object> initPrefs)
	{
		super.initialize(initPrefs);
		tempDir = new File((String) initPrefs.get(TEMP_DIRECTORY));
	}


	@Override
	public String getName()
	{
		reloadResourceBundle(TEXT_BUNDLE_PATH);
		return textsBundle.getString("ds-url.pluginName");
	}


	@Override
	public JComponent getPanel(FileSelectionListener listener)
	{
		if (panel == null) {
			panel = new URLInputPanel(this);
			reloadResourceBundle(TEXT_BUNDLE_PATH);
			panel.setLanguageBundle(textsBundle);
		}
		fileSelectionListener = listener;
		panel.clear();
		tempFile = null;
		return panel;
	}


	public void linksChanged(URL mainLink, List<URL> additionalLinks)
	{
		if (mainLink == null) {
			clearFileInfos();
			return;
		}
		if (tempFile == null) {
			try {
				tempFile = File.createTempFile("ds-url-", ".xml", tempDir);
			}
			catch (IOException e) {
				logger.error("Could not create temporary file", e);
				clearFileInfos();
				return;
			}
		}
		try {
			URLPublicationBuilder.saveURLPublicationFile(mainLink,
				additionalLinks, tempFile);
			FileInfo fileInfo = new SimpleFileInfo(tempFile.toURI().toURL(),
					true, "/" + URLPublicationBuilder.XML_FILE_NAME);
			fileSelectionListener.filesSelected(Arrays.asList(fileInfo));
		}
		catch (ParserConfigurationException e) {
			clearFileInfos();
			logger.error("XML generation failed", e);
		}
		catch (TransformerException e) {
			clearFileInfos();
			logger.error("XML generation failed", e);
		}
		catch (MalformedURLException e) {
			throw new RuntimeException("Local file path must be a valid url", e);
		}
		catch (LocalResourceUnavailableException e) {
			throw new RuntimeException("Local file must be available", e);
		}
	}


	private void clearFileInfos()
	{
		try {
			fileSelectionListener.filesSelected(null);
		}
		catch (LocalResourceUnavailableException e) {
			throw new RuntimeException("We don't post local resources", e);
		}
	}


	@Override
	public boolean isConfigurable()
	{
		return false;
	}


	@Override
	protected void reloadResourceBundle(String bundlePath)
	{
		if (textsBundle != null
				&& Locale.getDefault().equals(textsBundle.getLocale())) {
			return;
		}
		textsBundle = ResourceBundle.getBundle(bundlePath, Locale.getDefault());
		if (panel != null) {
			panel.setLanguageBundle(textsBundle);
		}
	}


	@Override
	public String toString()
	{
		return getName();
	}

}
