package pl.psnc.dlibra.app.extension.dsurl;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.ResourceBundle;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.UIManager;
import javax.swing.border.TitledBorder;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableModel;

import org.apache.log4j.Logger;

import pl.psnc.dlibra.app.common.component.inprogress.OperationInProgressManager;
import pl.psnc.dlibra.app.common.files.EditionFilesFactory;

/**
 * Main panel of the ds-url extension, used to input URLs the link
 * publication should point to.
 */
public class URLInputPanel
	extends JPanel
{

	/**
	 * Interface for objects that will be notified about changes in this panel.
	 */
	public static interface ChangesListener
	{

		public void linksChanged(URL mainLink, List<URL> additionalLinks);
	}

	private final static Logger logger = Logger.getLogger(URLInputPanel.class);

	private final static String LINK_STUB = "http://";

	private final ChangesListener listener;

	private ResourceBundle textsBundle;

	private JLabel mainLinkLabel;

	private JTextField mainLinkField;

	private JPanel additionalLinksPanel;

	private JTable additionalLinksTable;

	private JLabel summaryLabel;

	private JButton copyMainLinkButton;

	private JButton addLinkButton;

	private JButton removeLinksButton;

	private JButton addDerivedLinksButton;

	private boolean isUpdateScheduled = false;


	public URLInputPanel(ChangesListener dataSource)
	{
		this.listener = dataSource;

		layoutComponents();
		addListeners();
	}


	private void layoutComponents()
	{
		setLayout(new BorderLayout());
		JPanel mainLinkPanel = new JPanel(new BorderLayout());
		add(mainLinkPanel, BorderLayout.PAGE_START);
		mainLinkPanel.setBorder(BorderFactory.createEmptyBorder(15, 10, 10, 10));
		mainLinkPanel.add(mainLinkLabel = new JLabel(), BorderLayout.LINE_START);
		mainLinkPanel.add(mainLinkField = new JTextField(), BorderLayout.CENTER);

		additionalLinksPanel = new JPanel(new BorderLayout());
		add(additionalLinksPanel, BorderLayout.CENTER);
		additionalLinksPanel.setBorder(BorderFactory.createTitledBorder(""));
		additionalLinksPanel.add(new JScrollPane(
				prepareAdditionalLinksComponent()), BorderLayout.CENTER);
		additionalLinksPanel.add(summaryLabel = new JLabel(),
				BorderLayout.PAGE_END);
		JPanel rightPanel = new JPanel();
		additionalLinksPanel.add(rightPanel, BorderLayout.LINE_END);
		JPanel buttonsPanel = new JPanel(new GridLayout(4, 1, 5, 5));
		rightPanel.add(buttonsPanel);
		buttonsPanel.add(copyMainLinkButton = new JButton());
		buttonsPanel.add(addLinkButton = new JButton());
		buttonsPanel.add(removeLinksButton = new JButton());
		buttonsPanel.add(addDerivedLinksButton = new JButton());
	}


	private Component prepareAdditionalLinksComponent()
	{
		additionalLinksTable = new JTable(0, 1);
		additionalLinksTable.setTableHeader(null);
		additionalLinksTable.putClientProperty("terminateEditOnFocusLost",
			Boolean.TRUE);
		additionalLinksTable.getColumnModel().getColumn(0)
				.setCellRenderer(new DefaultTableCellRenderer() {

					@Override
					public Component getTableCellRendererComponent(
							JTable table, Object value, boolean isSelected,
							boolean hasFocus, int row, int column)
					{
						Component result = super.getTableCellRendererComponent(
							table, value, isSelected, hasFocus, row, column);
						setForeground(toURL(value.toString()) == null ? Color.red
								: Color.black);
						return result;
					}
				});

		final String DELETE = "Delete";
		additionalLinksTable.getInputMap(JComponent.WHEN_FOCUSED)
				.put(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0), DELETE);
		additionalLinksTable.getActionMap().put(DELETE, new AbstractAction() {

			public void actionPerformed(ActionEvent e)
			{
				removeSelectedLinks();
			}
		});

		return additionalLinksTable;
	}


	private void addListeners()
	{
		DocumentListener documentListener = new DocumentListener() {

			@Override
			public void removeUpdate(DocumentEvent e)
			{
				scheduleUpdate();
			}


			@Override
			public void insertUpdate(DocumentEvent e)
			{
				scheduleUpdate();
			}


			@Override
			public void changedUpdate(DocumentEvent e)
			{
				scheduleUpdate();
			}
		};
		mainLinkField.getDocument().addDocumentListener(documentListener);

		copyMainLinkButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e)
			{
				String mainLink = mainLinkField.getText();
				DefaultTableModel model = (DefaultTableModel) additionalLinksTable
						.getModel();
				model.addRow(new Object[] { mainLink});
				additionalLinksTable.getSelectionModel().setSelectionInterval(
					model.getRowCount() - 1, model.getRowCount() - 1);
			}
		});

		addLinkButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e)
			{
				DefaultTableModel model = (DefaultTableModel) additionalLinksTable
						.getModel();
				model.addRow(new Object[] { LINK_STUB});
				int lastRow = model.getRowCount() - 1;
				additionalLinksTable.getSelectionModel().setSelectionInterval(
					lastRow, lastRow);
				additionalLinksTable.editCellAt(lastRow, 0);
				additionalLinksTable.getEditorComponent().requestFocus();
			}
		});

		removeLinksButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e)
			{
				removeSelectedLinks();
			}

		});

		additionalLinksTable.getSelectionModel().addListSelectionListener(
			new ListSelectionListener() {

				@Override
				public void valueChanged(ListSelectionEvent e)
				{
					updateButtonsState();
				}
			});

		additionalLinksTable.getModel().addTableModelListener(
			new TableModelListener() {

				@Override
				public void tableChanged(TableModelEvent e)
				{
					scheduleUpdate();
				}
			});

		addDerivedLinksButton.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e)
			{
				ArrayList<String> selectedURLs = new ArrayList<String>();
				int[] selectedRows = additionalLinksTable.getSelectedRows();
				DefaultTableModel model = (DefaultTableModel) additionalLinksTable
						.getModel();
				for (int i = selectedRows.length - 1; i >= 0; i--) {
					String url = model.getValueAt(selectedRows[i], 0).toString();
					selectedURLs.add(url);
				}
				addDerivedLinks(selectedURLs);
			}
		});
	}


	private void scheduleUpdate()
	{
		if (isUpdateScheduled)
			return;
		isUpdateScheduled = true;
		SwingUtilities.invokeLater(new Runnable() {
			
			@Override
			public void run()
			{
				update();			
			}
		});
	}


	private void update()
	{
		isUpdateScheduled = false;
		URL mainLink = toURL(mainLinkField.getText());
		mainLinkField.setForeground(mainLink == null ? Color.red : Color.black);

		List<URL> additionalLinks = new ArrayList<URL>();
		Set<String> duplicatesDetector = new HashSet<String>(); // contains Strings to avoid using equals() on URLs.
		TableModel tableModel = additionalLinksTable.getModel();
		int rowCount = tableModel.getRowCount();
		for (int i = 0; i < rowCount; i++) {
			String value = tableModel.getValueAt(i, 0).toString();
			URL url = toURL(value);
			if (url != null && duplicatesDetector.add(value))
				additionalLinks.add(url);
		}

		listener.linksChanged(mainLink, additionalLinks);

		summaryLabel.setText(textsBundle.getString("additionalLinks.summary")
				.replace("{0}", "" + additionalLinks.size()));

		updateButtonsState();
	}


	private void updateButtonsState()
	{
		String mainLink = mainLinkField.getText();
		boolean isMainLinkAdded = false;
		TableModel tableModel = additionalLinksTable.getModel();
		int rowCount = tableModel.getRowCount();
		for (int i = 0; i < rowCount; i++) {
			if (mainLink.equals(tableModel.getValueAt(i, 0))) {
				isMainLinkAdded = true;
				break;
			}
		}
		setButtonEnabled(copyMainLinkButton, !isMainLinkAdded);

		int[] selectedRows = additionalLinksTable.getSelectedRows();
		setButtonEnabled(removeLinksButton, selectedRows.length > 0);
		setButtonEnabled(addDerivedLinksButton, selectedRows.length > 0);
	}


	/**
	 * Workaround for <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4783068">
	 * http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4783068</a> to make
	 * sure the text collor on disabled button is OK. 
	 */
	private void setButtonEnabled(JButton button, boolean enabled)
	{
		button.setEnabled(enabled);
		String text = button.getText();
		if (text.contains("<html>")) {
			Color color;
			if (enabled) {
				color = UIManager.getColor("Button.foreground");
			}
			else {
				color = UIManager.getColor("Button.disabledForeground");
				if (color == null)
					color = UIManager.getColor("Button.disabledText");
			}
			String fontTag = String.format("<font color=\"#%02X\">",
				color.getRGB() & 0xFFFFFF);
			if (text.contains("<font color=")) {
				text = text
						.replaceFirst("<font color=\"#[0-9A-F]+\">", fontTag);
			}
			else {
				text = text.replace("<html>", "<html>" + fontTag);
			}
			button.setText(text);
		}
	}


	private URL toURL(String url)
	{
		try {
			return new URI(url).toURL();
		}
		catch (Exception e) {
			return null;
		}
	}


	private void removeSelectedLinks()
	{
		int[] selectedRows = additionalLinksTable.getSelectedRows();
		additionalLinksTable.getSelectionModel().setSelectionInterval(-1, -1);
		DefaultTableModel model = (DefaultTableModel) additionalLinksTable
				.getModel();
		for (int i = selectedRows.length - 1; i >= 0; i--) {
			model.removeRow(selectedRows[i]);
		}
	}


	private void addDerivedLinks(final ArrayList<String> selectedURLs)
	{
		final OperationInProgressManager oipm = new OperationInProgressManager(
				SwingUtilities.getWindowAncestor(this));
		SwingWorker<List<URL>, ? > worker = new SwingWorker<List<URL>, String>() {
			private List<String> failed = new ArrayList<String>();

			@Override
			protected List<URL> doInBackground()
				throws Exception
			{
				oipm.start();
				List<URL> derived = new ArrayList<URL>();
				try {
					for (String urlString : selectedURLs) {
						oipm.setProgressInfo(urlString);
						try {
							URL url = new URL(urlString);
							List<URL> found = EditionFilesFactory.getInstance().findRelated(url);
							if (found == null) {
								failed.add(urlString);
								continue;
							}
							derived.addAll(found);
						}
						catch (Exception e) {
							logger.error(
								"Finding derived resources failed for "
										+ urlString, e);
							failed.add(urlString);
							continue;
						}
					}
				}
				finally {
					oipm.stop();
				}
				return derived;
			}


			@Override
			protected void done()
			{
				try {
					List<URL> derivedLinks = get();
					DefaultTableModel model = (DefaultTableModel) additionalLinksTable
							.getModel();
					for (URL url : derivedLinks) {
						model.addRow(new Object[] { url.toString()});
					}
				}
				catch (Exception e) {
					// ignore
				}

				if (!failed.isEmpty()) {
					StringBuilder message = new StringBuilder();
					message.append("<html>");
					message.append(textsBundle
							.getString("addDerivedLinks.failMessage"));
					message.append("<ul>");
					for (String url : failed) {
						message.append("<li>").append(url).append("</li>");
					}
					JOptionPane.showMessageDialog(URLInputPanel.this, message,
						textsBundle.getString("addDerivedLinks.failTitle"),
						JOptionPane.WARNING_MESSAGE);
				}
			}
		};
		worker.execute();
	}


	/**
	 * Clears all data currently entered into this panel and prepares
	 * it for a new session. 
	 */
	public void clear()
	{
		mainLinkField.setText(LINK_STUB);
		DefaultTableModel model = (DefaultTableModel) additionalLinksTable
				.getModel();
		model.setRowCount(0);
		updateButtonsState();
	}


	public void setLanguageBundle(ResourceBundle textsBundle)
	{
		this.textsBundle = textsBundle;
		mainLinkLabel.setText(textsBundle.getString("mainLink.label"));
		mainLinkLabel.setToolTipText(textsBundle.getString("mainLink.tooltip"));
		mainLinkField.setToolTipText(textsBundle.getString("mainLink.tooltip"));
		((TitledBorder) additionalLinksPanel.getBorder()).setTitle(textsBundle
				.getString("additionalLinks.label"));
		additionalLinksTable.setToolTipText(textsBundle
				.getString("additionalLinks.tooltip"));

		copyMainLinkButton.setText(textsBundle.getString("copyMainLink.label"));
		copyMainLinkButton.setToolTipText(textsBundle
				.getString("copyMainLink.tooltip"));
		addLinkButton.setText(textsBundle.getString("addLink.label"));
		removeLinksButton.setText(textsBundle.getString("removeLinks.label"));
		addDerivedLinksButton.setText(textsBundle
				.getString("addDerivedLinks.label"));
		addDerivedLinksButton.setToolTipText(textsBundle
				.getString("addDerivedLinks.tooltip"));
	}


	public void setContents(URL mainLink, ArrayList<URL> additionalLinks)
	{
		isUpdateScheduled = true; // block updating
		mainLinkField.setText(mainLink.toString());
		DefaultTableModel model = (DefaultTableModel) additionalLinksTable
				.getModel();
		model.setRowCount(0);
		for (URL url : additionalLinks) {
			model.addRow(new Object[] { url.toString()});
		}
		isUpdateScheduled = false;
		additionalLinksTable.getSelectionModel().setSelectionInterval(-1, -1);
		summaryLabel.setText(textsBundle.getString("additionalLinks.summary")
			.replace("{0}", "" + additionalLinks.size()));
	}
}
