package pl.psnc.dlibra.app.extension.dsurl;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.List;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class URLPublicationBuilder
{
	public static final String XML_FILE_NAME = "url-publication.xml";

	private static final String ROOT_ELEMENT = "url-publication";

	private static final String LINK_LIST_ELEMENT = "index-links";

	private static final String LINK_ELEMENT = "link";

	private static final String MAIN_LINK_ATTRIBUTE = "mainLink";

	private static final String LINK_ATTRIBUTE = "url";


	private URLPublicationBuilder()
	{
		//hide constructor
	}


	public static void saveURLPublicationFile(URL mainURL,
			List<URL> urlsToIndex, File targetFile)
		throws ParserConfigurationException, TransformerException
	{
		Document doc = DocumentBuilderFactory.newInstance()
				.newDocumentBuilder().newDocument();
		Element root = doc.createElement(ROOT_ELEMENT);
		root.setAttribute(MAIN_LINK_ATTRIBUTE, mainURL.toString());
		doc.appendChild(root);

		if (!urlsToIndex.isEmpty()) {
			Element otherList = doc.createElement(LINK_LIST_ELEMENT);
			root.appendChild(otherList);

			for (URL url : urlsToIndex) {
				Element link = doc.createElement(LINK_ELEMENT);
				link.setAttribute(LINK_ATTRIBUTE, url.toString());
				otherList.appendChild(link);
			}
		}

		Transformer transformer = TransformerFactory.newInstance()
				.newTransformer();
		transformer.setOutputProperty(OutputKeys.INDENT, "yes");
		transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
		DOMSource source = new DOMSource(doc);
		StreamResult result = new StreamResult(targetFile);
		transformer.transform(source, result);
	}


	public static URL loadURLPublicationFile(File sourceFile,
			List<URL> outLinskToIndex)
		throws SAXException, IOException, ParserConfigurationException
	{
		Document document = DocumentBuilderFactory.newInstance()
				.newDocumentBuilder().parse(sourceFile);
		return parseURLPublicationDoc(document, outLinskToIndex);
	}


	public static URL loadURLPublicationStream(InputStream input,
			List<URL> outLinskToIndex)
		throws SAXException, IOException, ParserConfigurationException
	{
		Document document = DocumentBuilderFactory.newInstance()
				.newDocumentBuilder().parse(input);
		return parseURLPublicationDoc(document, outLinskToIndex);
	}


	private static URL parseURLPublicationDoc(Document document,
			List<URL> outLinskToIndex)
		throws IOException, MalformedURLException
	{
		NodeList list = document.getElementsByTagName(ROOT_ELEMENT);
		if (list.getLength() != 1) {
			throw new IOException("There must exactly one " + ROOT_ELEMENT
					+ " element.");
		}
		String mainLink = list.item(0).getAttributes()
				.getNamedItem(MAIN_LINK_ATTRIBUTE).getTextContent();

		if (outLinskToIndex != null) {
			list = document.getElementsByTagName(LINK_ELEMENT);
			for (int i = 0; i < list.getLength(); i++) {
				String url = list.item(i).getAttributes()
						.getNamedItem(LINK_ATTRIBUTE).getTextContent();
				outLinskToIndex.add(new URL(url));
			}
		}
		return (new URL(mainLink));
	}
}
