package pl.psnc.dlibra.app.extension.dsurl;

import java.awt.BorderLayout;
import java.awt.Window;
import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.swing.JPanel;

import pl.psnc.dlibra.app.ExceptionHandler;
import pl.psnc.dlibra.app.common.component.extension.AbstractConfigurableExtension;
import pl.psnc.dlibra.app.common.component.inprogress.OperationInProgressManager;
import pl.psnc.dlibra.app.extension.dsurl.URLInputPanel.ChangesListener;
import pl.psnc.dlibra.app.extension.fileshandler.ApplyCancelInformation;
import pl.psnc.dlibra.app.extension.fileshandler.FilesHandler;
import pl.psnc.dlibra.app.extension.fileshandler.UnsupportedSetOfFilesException;
import pl.psnc.dlibra.app.gui.common.aclpanel.ApplyCancelPanel;

public class URLPublicationHandler
	extends AbstractConfigurableExtension
	implements FilesHandler, ChangesListener
{

	public class ApplyCancelListener
	implements pl.psnc.dlibra.app.gui.base.ChangesListener
	{
	
		private File fileToEdit;
	
		private ApplyCancelInformation aci;
	
	
		public ApplyCancelListener(File fileToEdit, ApplyCancelInformation aci)
		{
			this.fileToEdit = fileToEdit;
			this.aci = aci;
		}
	
	
		@Override
		public void applyChanges(OperationInProgressManager manager)
			throws Exception
		{
			URLPublicationBuilder.saveURLPublicationFile(mainLink, additionalLinks, fileToEdit);
			aci.applied();
			applyCancel.setApplyButtonEnabled(false);
		}
	
	
		@Override
		public void cancelChanges()
		{
			try {
				setContents(fileToEdit);
				aci.cancelled();
				applyCancel.setApplyButtonEnabled(false);
			}
			catch (Exception e) {
				ExceptionHandler.getInstance().handleException(e, urlInputPanel);
			}
		}
	
	
		@Override
		public void contentChanged()
		{
			// ignore
		}
	}
	
	private URLInputPanel urlInputPanel;

	private ApplyCancelPanel applyCancel;

	private URL mainLink;

	private List<URL> additionalLinks;


	@Override
	public List<File> handleFiles(Window parentWindow, List<File> filesToHandle)
		throws UnsupportedSetOfFilesException
	{
		// URL publication file does not need special treating
		throw new UnsupportedSetOfFilesException();
	}


	@Override
	public boolean editFile(JPanel panel, File fileToEdit,
			ApplyCancelInformation aci)
		throws UnsupportedSetOfFilesException
	{
		if (urlInputPanel == null)
			urlInputPanel = new URLInputPanel(this);
		panel.setLayout(new BorderLayout());
		panel.add(urlInputPanel, BorderLayout.CENTER);

		applyCancel = new ApplyCancelPanel(panel);
		applyCancel.setTarget(new ApplyCancelListener(fileToEdit, aci));
		applyCancel.setApplyButtonEnabled(false);
		panel.add(applyCancel, BorderLayout.PAGE_END);

		reloadResourceBundle(URLDataSource.TEXT_BUNDLE_PATH);

		return setContents(fileToEdit);
	}


	private boolean setContents(File fileToEdit)
	{
		URL mainLink;
		ArrayList<URL> additionalLinks = new ArrayList<URL>();
		try {
			mainLink = URLPublicationBuilder.loadURLPublicationFile(fileToEdit, additionalLinks);
		}
		catch (Exception e) {
			ExceptionHandler.getInstance().handleException(e, urlInputPanel);
			return false;
		}
		urlInputPanel.setContents(mainLink, additionalLinks);
		return true;
	}


	@Override
	public boolean supportsEdit(String fileName)
	{
		// startsWith() instead of equals(), because given string
		// has file modification date added at the end.
		return fileName.startsWith(URLPublicationBuilder.XML_FILE_NAME);
	}


	@Override
	public void linksChanged(URL mainLink, List<URL> additionalLinks)
	{
		this.mainLink = mainLink;
		this.additionalLinks = additionalLinks;
		applyCancel.setApplyButtonEnabled(true);
	}


	@Override
	public boolean isConfigurable()
	{
		return false;
	}


	@Override
	protected void reloadResourceBundle(String bundlePath)
	{
		if (textsBundle != null
				&& Locale.getDefault().equals(textsBundle.getLocale())) {
			return;
		}
		textsBundle = ResourceBundle.getBundle(bundlePath, Locale.getDefault());
		if (urlInputPanel != null) {
			urlInputPanel.setLanguageBundle(textsBundle);
		}
	}


	@Override
	public String toString()
	{
		return "Link publicatoin handler";
	}
}
