/*
 * Copyright (c) 2010 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.app.extension.fh.jpg;

import java.awt.Image;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Map;

import javax.imageio.ImageIO;

import pl.psnc.dlibra.app.extension.miniatureprovider.MiniatureProvider;


public class JPGFHMiniatureProvider
	implements MiniatureProvider
{

	private List<FileInfo> fileInfos;

	private String parentDir;


	@Override
	public void initialize(Map<String, Object> initPrefs)
	{
		// nothing to initialize
	}


	@Override
	public boolean setMainFile(File mainFile)
	{
		if (ListOfObjectsEditor.PRESENTATION_DATA_FILENAME.equals(mainFile
				.getName())) {
			try {
				fileInfos = new XMLDescriptionFileParser(mainFile.toURI().toURL()).parse();
				parentDir = mainFile.getParent();
			}
			catch (Exception e) {
				// invalid xml file
				return false;
			}
			return !fileInfos.isEmpty();
		}
		return false;
	}


	@Override
	public int countMiniatures()
	{
		return fileInfos.size();
	}


	@Override
	public Image getMiniature(int index)
		throws IOException
	{
		File file = new File(fileInfos.get(index).getFilePath());
		if (!file.isAbsolute()) {
			file = new File(parentDir, file.getPath());
		}
		return ImageIO.read(file);
	}


	@Override
	public void dispose() {
		fileInfos = null;
	}


	@Override
	public MiniatureProvider clone()
		throws CloneNotSupportedException
	{
		return (MiniatureProvider) super.clone();
	}


	@Override
	public String toString()
	{
		return "Miniature provider for multi-JPG publications";
	}

}
