/**
 * $Id: JPGFilesHandler.java,v 1.3 2009-04-30 06:51:04 tparkola Exp $
 *
 * Copyright (c) 2002 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.app.extension.fh.jpg;

import java.awt.Window;
import java.awt.Dialog.ModalityType;
import java.io.File;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;

import javax.activation.MimetypesFileTypeMap;
import javax.swing.JPanel;

import pl.psnc.dlibra.app.extension.Extension;
import pl.psnc.dlibra.app.extension.fileshandler.ApplyCancelInformation;
import pl.psnc.dlibra.app.extension.fileshandler.FilesHandler;
import pl.psnc.dlibra.app.extension.fileshandler.UnsupportedSetOfFilesException;

/**
 * File parser for djvu files
 */
public class JPGFilesHandler
	implements FilesHandler, Extension
{

	private static final String DEFAULT_BUNDLE_PATH = "pl/psnc/dlibra/app/extension/fh/jpg/res/GUI";

	private static final String[] SUPPORTED_MIMETYPES = new String[] {
			"image/jpeg", "image/gif", "image/png"};

	private static final String[] IGNORED_FILES = new String[] { "thumb.db",
			"ehthumbs.db"};

	private MimetypesFileTypeMap mimetypesFileTypeMap = new MimetypesFileTypeMap(
			this.getClass().getResourceAsStream(
				"/pl/psnc/dlibra/app/extension/fh/jpg/file-types.properties"));

	private ResourceBundle textsBundle;

	private SupportChecker<String> supportChecker;


	public JPGFilesHandler()
	{
		supportChecker = new SupportChecker<String>(SUPPORTED_MIMETYPES,
				mimetypesFileTypeMap, IGNORED_FILES);
		reloadResourceBundle(DEFAULT_BUNDLE_PATH);
	}


	public void initialize(Map<String, Object> initPrefs)
	{
		// no information is needed
	}


	/**
	 * Reloads resources using given path to bundles.
	 * @param bundlePath
	 */
	private void reloadResourceBundle(String bundlePath)
	{
		textsBundle = ResourceBundle.getBundle(bundlePath, Locale.getDefault());
	}


	@Override
	public String toString()
	{
		return "JPG files handler";
	}


	public List<File> handleFiles(Window parentWindow, List<File> filesToHandle)
		throws UnsupportedSetOfFilesException
	{
		reloadResourceBundle(DEFAULT_BUNDLE_PATH);

		if (!supportChecker.supports(extractNames(filesToHandle))) {
			throw new UnsupportedSetOfFilesException();
		}

		// remove files that do not constitute the content
		// and determine whether the meta file exists
		Iterator<File> filesIter = filesToHandle.iterator();
		File metaFile = null;
		while (filesIter.hasNext()) {
			File file = filesIter.next();
			for (String ign : IGNORED_FILES) {
				if (ign.equalsIgnoreCase(file.getName())) {
					filesIter.remove();
				}
			}

			if ((ListOfObjectsEditor.PRESENTATION_DATA_FILENAME.equals(file
					.getName()))) {
				metaFile = file;
				filesIter.remove();
			}
		}

		// preliminary sort
		Collections.sort(filesToHandle, new Comparator<File>() {

			@Override
			public int compare(File o1, File o2)
			{
				return Collator.getInstance().compare(o1.getName(),
					o2.getName());
			}
		});

		// show the results of processing to the user
		ListOfObjectsEditor listOfObjectsEditor = new ListOfObjectsEditor(
				parentWindow, ModalityType.DOCUMENT_MODAL);
		listOfObjectsEditor.setResourceBundle(textsBundle);
		listOfObjectsEditor.pack();
		if (metaFile != null) {
			// prepare file infos
			try {
				List<FileInfo> infos = new ArrayList<FileInfo>();
				XMLDescriptionFileParser parser = new XMLDescriptionFileParser(
						metaFile.toURI().toURL());
				for (FileInfo fi : parser.parse()) {
					String filePath = metaFile.getParent() + File.separator + fi.toString();
					File file = new File(filePath);
					if (file.exists()) {
						infos.add(fi);
					}
				}

				listOfObjectsEditor.setFileInfos(infos, metaFile.getParentFile());
			}
			catch (Exception e) {
				// try normal way
				listOfObjectsEditor.setObjects(filesToHandle);
			}
		}
		else {
			listOfObjectsEditor.setObjects(filesToHandle);
		}

		listOfObjectsEditor.setLocationRelativeTo(parentWindow);
		listOfObjectsEditor.setVisible(true);

		if (listOfObjectsEditor.getObjects() == null) {
			throw new UnsupportedSetOfFilesException();
		}

		return listOfObjectsEditor.getObjects();
	}


	private List<String> extractNames(List<File> filesToHandle)
	{
		List<String> fileNames = new ArrayList<String>();

		for (File f : filesToHandle) {
			fileNames.add(f.getName());
		}
		return fileNames;
	}


	public boolean editFile(JPanel panel, File fileToEdit,
			ApplyCancelInformation api)
		throws UnsupportedSetOfFilesException
	{
		reloadResourceBundle(DEFAULT_BUNDLE_PATH);

		ListOfObjectsEditor listOfObjectsEditor = new ListOfObjectsEditor();
		listOfObjectsEditor.setResourceBundle(textsBundle);
		listOfObjectsEditor.pack();
		try {
			listOfObjectsEditor.setObject(fileToEdit, api);
		}
		catch (Exception e) {
			throw new UnsupportedSetOfFilesException();
		}

		listOfObjectsEditor.initializeContainer(panel);

		return true;
	}


	public boolean supportsEdit(String fileName)
	{
		return fileName.startsWith(ListOfObjectsEditor.PRESENTATION_DATA_FILENAME);
	}


	public boolean supportsFiles(List<String> files)
	{
		return supportChecker.supports(files);
	}
}
