package pl.psnc.dlibra.app.extension.fh.jpg;

import java.awt.Container;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;

import javax.swing.BorderFactory;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.xml.sax.SAXException;

import pl.psnc.dlibra.app.extension.fileshandler.ApplyCancelInformation;

public class ListOfObjectsEditor
	extends JDialog
{

	public static final String PRESENTATION_DATA_FILENAME = "PresentationData.xml";

	private JList listOfObjects;

	private FileInfoEditingPanel editor;

	private List<File> results = null;

	private File mainDir;

	private JPanel mainPanel;

	private JButton upButton;

	private JButton downButton;

	private JButton applyButton;

	private JButton cancelButton;

	private ApplyCancelInformation api;

	private ResourceBundle textsBundle;

	private File fileToEdit;


	public ListOfObjectsEditor(Window parent, ModalityType type)
	{
		super(parent, type);

		initialize();

		initializeContainer(this);

		initializeListeners();
	}


	public ListOfObjectsEditor()
	{
		super();

		initialize();

		initializeListeners();
	}


	private void initializeListeners()
	{
		listOfObjects.addListSelectionListener(new ListSelectionListener() {

			@Override
			public void valueChanged(ListSelectionEvent e)
			{
				editor.applyChanges();
				editor.setObject((FileInfo) listOfObjects.getSelectedValue());
			}
		});
	}


	private void initializeGUITexts(ResourceBundle textsBundle)
	{
		setTitle(textsBundle.getString("editor.title"));
		mainPanel.setBorder(BorderFactory.createTitledBorder(textsBundle
				.getString("mainPanel.title")));
		upButton.setText(textsBundle.getString("upButton.text"));
		downButton.setText(textsBundle.getString("downButton.text"));
		editor.setResourceBundle(textsBundle);
		if (applyButton != null) {
			applyButton.setText(textsBundle.getString("applyButton.text"));
		}
		if (cancelButton != null) {
			cancelButton.setText(textsBundle.getString("cancelButton.text"));
		}
	}


	private void initialize()
	{
		mainPanel = new JPanel();

		mainPanel.setLayout(new GridBagLayout());

		mainPanel
				.setBorder(BorderFactory.createTitledBorder("list of objects"));

		GridBagConstraints c = new GridBagConstraints();
		c.weightx = 1;
		c.gridwidth = 2;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.insets = new Insets(10, 10, 5, 5);
		//		mainPanel.add(listLabel, c);

		c = new GridBagConstraints();
		c.weighty = 1;
		c.weightx = 1;
		c.gridy = 1;
		c.gridwidth = 2;
		c.insets = new Insets(15, 10, 5, 5);
		c.fill = GridBagConstraints.BOTH;
		listOfObjects = new JList(new DefaultListModel());
		mainPanel.add(new JScrollPane(listOfObjects,
				JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
				JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS), c);

		c = new GridBagConstraints();
		c.weightx = 0.3;
		c.gridy = 2;
		c.insets = new Insets(0, 10, 10, 2);
		c.fill = GridBagConstraints.HORIZONTAL;
		upButton = new JButton("up");
		upButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e)
			{
				int selected = listOfObjects.getSelectedIndex();
				int newIndex = listOfObjects.getSelectedIndex() - 1;

				if (newIndex < 0) {
					return;
				}

				Object valueSelected = listOfObjects.getSelectedValue();
				Object toReplace = ((DefaultListModel) listOfObjects.getModel())
						.elementAt(newIndex);

				((DefaultListModel) listOfObjects.getModel()).setElementAt(
					valueSelected, newIndex);
				((DefaultListModel) listOfObjects.getModel()).setElementAt(
					toReplace, selected);

				listOfObjects.setSelectedValue(valueSelected, true);
			}
		});
		mainPanel.add(upButton, c);

		c = new GridBagConstraints();
		c.weightx = 0.3;
		c.gridy = 2;
		c.gridx = 1;
		c.insets = new Insets(0, 0, 10, 5);
		c.fill = GridBagConstraints.HORIZONTAL;
		downButton = new JButton("down");
		downButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e)
			{
				int selected = listOfObjects.getSelectedIndex();
				int newIndex = listOfObjects.getSelectedIndex() + 1;

				if (newIndex >= listOfObjects.getModel().getSize()) {
					return;
				}

				Object valueSelected = listOfObjects.getSelectedValue();
				Object toReplace = ((DefaultListModel) listOfObjects.getModel())
						.elementAt(newIndex);

				((DefaultListModel) listOfObjects.getModel()).setElementAt(
					valueSelected, newIndex);
				((DefaultListModel) listOfObjects.getModel()).setElementAt(
					toReplace, selected);

				listOfObjects.setSelectedValue(valueSelected, true);
			}
		});
		mainPanel.add(downButton, c);

		c = new GridBagConstraints();
		c.weightx = 1;
		c.weighty = 1;
		c.gridheight = 2;
		c.gridwidth = 2;
		c.gridy = 0;
		c.gridx = 2;
		c.fill = GridBagConstraints.BOTH;
		c.insets = new Insets(10, 0, 10, 10);
		editor = new FileInfoEditingPanel();
		mainPanel.add(editor, c);
	}


	public void initializeContainer(Container cmp)
	{
		GridBagConstraints c;
		cmp.setLayout(new GridBagLayout());

		c = new GridBagConstraints();
		c.fill = GridBagConstraints.BOTH;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 2;
		cmp.add(mainPanel, c);

		c = new GridBagConstraints();
		c.weightx = 1;
		c.anchor = GridBagConstraints.EAST;
		c.gridy = 1;
		c.gridx = 0;
		c.insets = new Insets(10, 10, 10, 3);
		applyButton = new JButton("apply");
		applyButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e)
			{
				editor.applyChanges();

				List<FileInfo> infos = new ArrayList<FileInfo>();

				for (int i = 0; i < ((DefaultListModel) listOfObjects
						.getModel()).getSize(); i++) {
					infos.add((FileInfo) ((DefaultListModel) listOfObjects
							.getModel()).get(i));
				}

				XMLDescriptionFileBuilder builder = new XMLDescriptionFileBuilder(
						infos);

				File mainFile = fileToEdit == null ? new File(mainDir,
						PRESENTATION_DATA_FILENAME) : fileToEdit;

				try {
					builder.storeXML(mainFile);
				}
				catch (ParserConfigurationException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
				catch (TransformerException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
				catch (FileNotFoundException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}

				List<File> objectFiles = new ArrayList<File>();

				if (mainFile.exists()) {
					objectFiles.add(mainFile);
				}

				for (FileInfo fi : infos) {
					objectFiles.add(new File(mainDir, fi.getFilePath()));
				}

				results = objectFiles;

				if (api != null) {
					api.applied();
				}

				setVisible(false);
			}
		});

		cmp.add(applyButton, c);

		c = new GridBagConstraints();
		c.anchor = GridBagConstraints.EAST;
		c.gridy = 1;
		c.gridx = 1;
		c.insets = new Insets(10, 0, 10, 2);
		cancelButton = new JButton("cancel");
		cancelButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e)
			{
				results = null;

				if (api != null) {
					api.cancelled();
				}

				setVisible(false);
			}
		});
		cmp.add(cancelButton, c);

		if (textsBundle != null) {
			initializeGUITexts(textsBundle);
		}
	}


	@Override
	public void setVisible(boolean b)
	{
		if (b) {
			results = null;
		}
		super.setVisible(b);
	}


	public void setObjects(List<File> filesToHandle)
	{
		if (filesToHandle.size() > 0) {
			mainDir = filesToHandle.get(0).getParentFile();
		}
		final int mainPathLength = mainDir.getAbsolutePath().length();
		for (File f : filesToHandle) {
			String path = f.getAbsolutePath().substring(mainPathLength + 1);
			((DefaultListModel) listOfObjects.getModel())
					.addElement(new FileInfo(path));
		}
	}


	public void setFileInfos(List<FileInfo> filesToHandle, File mainDir)
	{
		this.mainDir = mainDir;

		for (FileInfo f : filesToHandle) {
			((DefaultListModel) listOfObjects.getModel()).addElement(f);
		}
	}


	public List<File> getObjects()
	{
		return results;
	}


	public void setResourceBundle(ResourceBundle textsBundle)
	{
		this.textsBundle = textsBundle;
		initializeGUITexts(textsBundle);
	}


	public void setObject(File fileToEdit, ApplyCancelInformation api)
		throws SAXException, IOException, ParserConfigurationException
	{
		this.fileToEdit = fileToEdit;
		this.api = api;

		XMLDescriptionFileParser parser = new XMLDescriptionFileParser(
				fileToEdit.toURI().toURL());

		setFileInfos(parser.parse(), fileToEdit.getParentFile());
	}
}
