/*
 * Copyright (c) 2000 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.app.extension.fh.jpg;

import java.util.List;

import javax.activation.MimetypesFileTypeMap;

public class SupportChecker<T>
{

	private final String[] supportedMimeTypes;

	private final MimetypesFileTypeMap mimetypesFileTypeMap;

	private final String[] ignoredFiles;


	public SupportChecker(String[] supportedMimeTypes,
			MimetypesFileTypeMap mimetypesFileTypeMap, String[] ignoredFiles)
	{
		this.supportedMimeTypes = supportedMimeTypes;
		this.mimetypesFileTypeMap = mimetypesFileTypeMap;
		this.ignoredFiles = ignoredFiles;
	}


	public boolean supports(List<T> elements)
	{
		MimeTypeRecognizer<T> recognizer = new MimeTypeRecognizer<T>(
				mimetypesFileTypeMap);

		for (T f : elements) {
			boolean supports = false;

			for (String type : supportedMimeTypes) {
				if (type.equalsIgnoreCase(recognizer.recognizeContentType(f))) {
					supports = true;
				}
			}

			// do not consider files that should be ignored
			for (String ignoredFile : ignoredFiles) {
				if (f != null && ignoredFile.equalsIgnoreCase(f.toString())) {
					supports = true;
				}
			}

			// ignore presentation data xml file if exists
			if (ListOfObjectsEditor.PRESENTATION_DATA_FILENAME.equals(f)) {
				supports = true;
			}

			if (!supports) {
				return false;
			}
		}

		return true;
	}
}
