package pl.psnc.dlibra.app.extension.fh.jpg;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.util.List;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.CDATASection;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class XMLDescriptionFileBuilder
{

	private final List<FileInfo> infos;


	public XMLDescriptionFileBuilder(List<FileInfo> infos)
	{
		this.infos = infos;
	}


	public void storeXML(File mainFile)
		throws ParserConfigurationException, TransformerException,
		FileNotFoundException
	{
		Document document = DocumentBuilderFactory.newInstance()
				.newDocumentBuilder().newDocument();

		Element root = document.createElement("object-presentation");

		document.appendChild(root);

		Element preElems = document.createElement("presentation-elements");

		root.appendChild(preElems);

		int counter = 0;
		for (FileInfo fi : infos) {
			preElems.appendChild(createElement(document, fi, counter++));
		}

		TransformerFactory factory = TransformerFactory.newInstance();

		Transformer t = factory.newTransformer();
		t.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
		t.setOutputProperty(OutputKeys.INDENT, "yes");

		DOMSource source = new DOMSource(document);

		StreamResult result = new StreamResult(new FileOutputStream(mainFile));

		t.transform(source, result);
	}


	private Node createElement(Document document, FileInfo fi, int counter)
	{
		Element element = document.createElement("presentation-element");
		element.setAttribute("position", String.valueOf(counter));

		Element fullImg = document.createElement("full-image");
		CDATASection data = document.createCDATASection(fi.getFilePath());
		fullImg.appendChild(data);
		element.appendChild(fullImg);

		if (!"".equals(fi.getDescription())) {
			Element description = document.createElement("element-hint");
			data = document.createCDATASection(fi.getDescription());
			description.appendChild(data);
			element.appendChild(description);
		}

		if (!"".equals(fi.getComment())) {
			Element comment = document.createElement("element-text");
			data = document.createCDATASection(fi.getComment());
			comment.appendChild(data);
			element.appendChild(comment);
		}

		return element;
	}
}
