package pl.psnc.dlibra.app.extension.fh.jpg;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class XMLDescriptionFileParser
{

	private final URL fileURL;


	public XMLDescriptionFileParser(URL fileURL)
	{
		this.fileURL = fileURL;
	}


	public List<FileInfo> parse()
		throws SAXException, IOException, ParserConfigurationException
	{
		InputStream stream = fileURL.openStream();
		return parse(stream);
	}


	public static List<FileInfo> parse(File fileToEdit)
		throws SAXException, IOException, ParserConfigurationException
	{
		Document document = DocumentBuilderFactory.newInstance()
				.newDocumentBuilder().parse(fileToEdit);
		return analyze(document);
	}


	public static List<FileInfo> parse(InputStream inputStream)
		throws SAXException, IOException, ParserConfigurationException
	{
		Document document = DocumentBuilderFactory.newInstance()
				.newDocumentBuilder().parse(inputStream);
		return analyze(document);
	}


	private static List<FileInfo> analyze(Document document)
	{
		NodeList list = document.getElementsByTagName("presentation-element");

		List<FileInfo> result = new ArrayList<FileInfo>();

		for (int i = 0; i < list.getLength(); i++) {
			FileInfo fi = prepareFile(list.item(i));
			if (fi != null) {
				result.add(fi);
			}
		}

		Collections.sort(result, new Comparator<FileInfo>() {

			@Override
			public int compare(FileInfo o1, FileInfo o2)
			{
				return o1.getPosition().compareTo(o2.getPosition());
			}
		});

		return result;
	}


	private static FileInfo prepareFile(Node item)
	{
		if (Node.ELEMENT_NODE == item.getNodeType()) {
			Element e = (Element) item;
			NodeList children = e.getChildNodes();
			String fullName = null;
			String hint = "";
			String text = "";
			for (int i = 0; i < children.getLength(); i++) {
				Node child = children.item(i);
				if ("full-image".equals(child.getNodeName())) {
					if (child.getChildNodes().getLength() > 0) {
						fullName = child.getChildNodes().item(0)
								.getTextContent();
					}
				}
				if ("element-hint".equals(child.getNodeName())) {
					if (child.getChildNodes().getLength() > 0) {
						hint = child.getChildNodes().item(0).getTextContent();
					}
				}
				if ("element-text".equals(child.getNodeName())) {
					if (child.getChildNodes().getLength() > 0) {
						text = child.getChildNodes().item(0).getTextContent();
					}
				}
			}
			if (fullName != null) {
				FileInfo fileInfo = new FileInfo(fullName);
				fileInfo.setComment(text);
				fileInfo.setDescription(hint);
				fileInfo.setPosition(Integer.valueOf(e.getAttribute("position")));
				return fileInfo;
			}
		}

		return null;
	}

}
