/*
 * Copyright (c) 2002 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.app.extension.dir;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Queue;

import pl.psnc.dlibra.app.extension.Extension;
import pl.psnc.dlibra.app.extension.fileparser.FileParser;

/**
 * File parser for directories - retrieves all of the files inside
 */
public class DirectoryFileParser
	implements FileParser, Extension
{

	public List<URL> parseFile(URL fileURL)
		throws IOException
	{
		try {
			ArrayList<URL> result = new ArrayList<URL>();
			File folder = new File(fileURL.toURI());
			List<File> files = scanForFiles(folder);
			for (File file : files) {
				result.add(file.toURI().toURL());
			}
			return result;
		}
		catch (URISyntaxException e) {
			throw new IllegalArgumentException(fileURL
					+ " is not a valid folder URL");
		}
	}


	public boolean supports(URL file)
	{
		try {
			return new File(file.toURI()).isDirectory();
		}
		catch (URISyntaxException e) {
			return false;
		}
		catch (IllegalArgumentException e) {
			return false;
		}
	}


	public void initialize(Map<String, Object> initPrefs)
	{
		// no information is needed
	}


	@Override
	public String toString()
	{
		return "Directory parser";
	}


	private List<File> scanForFiles(File rootFolder)
	{
		ArrayList<File> result = new ArrayList<File>();
		Queue<File> queue = new ArrayDeque<File>();
		queue.add(rootFolder);
		while (!queue.isEmpty()) {
			File[] files = queue.poll().listFiles();
			for (File file : files) {
				if (file.isDirectory()) {
					queue.add(file);
				}
				else {
					result.add(file);
				}
			}
		}
		return result;
	}
}
