/**
 * $Id: DjVuFileParser.java,v 1.3 2009-04-22 08:45:43 tparkola Exp $
 *
 * Copyright (c) 2002 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.app.extension.djvu;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Vector;
import java.util.regex.Pattern;

import javax.activation.MimetypesFileTypeMap;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import pl.psnc.dlibra.app.extension.Extension;
import pl.psnc.dlibra.app.extension.fileparser.FileParser;

import com.lizardtech.djvu.DjVmDir;
import com.lizardtech.djvu.Document;

/**
 * File parser for djvu files
 */
public class DjVuFileParser
	implements FileParser, Extension
{

	private static final String SUPPORTED_MIMETYPE = "image/x.djvu"; //$NON-NLS-1$

	private static final String DEFAULT_BUNDLE_PATH = "pl/psnc/dlibra/app/extension/djvu/res/GUI";

	private static final Pattern RECOMMENDED_FILENAME_PATTERN = Pattern
			.compile("[a-zA-Z0-9_\\-\\.]+");

	private MimetypesFileTypeMap mimetypesFileTypeMap = new MimetypesFileTypeMap(
			this.getClass().getResourceAsStream(
				"/pl/psnc/dlibra/app/extension/djvu/file-types.properties")); //$NON-NLS-1$


	public List<URL> parseFile(URL file)
		throws IOException
	{
		List<URL> result = new ArrayList<URL>();

		Document d;
		try {
			d = new Document(file);
		}
		catch (IOException e) {
			return result;
		}

		String path = file.getPath();
		String fileName = new File(path).getName();
		String parentPath = path.substring(0, path.length() - fileName.length());
		URL parentURL = new URL(file, parentPath);

		DjVmDir dir = d.getDjVmDir();
		if (dir.is_indirect()) {
			@SuppressWarnings("unchecked")
			Vector<DjVmDir.File> files = dir.get_files_list();
			for (DjVmDir.File dFile : files) {
				String filePath = dFile.get_load_name();
				// workaround: sometimes filePath is absolute
				while (filePath.startsWith("/") || filePath.startsWith("\\")) {
					filePath = filePath.substring(1);
				}
				result.add(new URL(parentURL, filePath));
			}
		}

		// show warning if any file name (including main file) contains potentially problematic characters
		result.add(file);
		checkFilename(result);
		result.remove(result.size() - 1);

		return result;
	}


	private void checkFilename(List<URL> files)
	{
		for (URL file : files) {
			String fileName = new File(file.getPath()).getName();
			if (!RECOMMENDED_FILENAME_PATTERN.matcher(fileName).matches()) {
				if (SwingUtilities.isEventDispatchThread()) {
					showSpacesWarning();
				}
				else {
					try {
						SwingUtilities.invokeAndWait(new Runnable() {

							public void run()
							{
								showSpacesWarning();
							}
						});
					}
					catch (InterruptedException e) {
						// ignore
					}
					catch (InvocationTargetException e) {
						// ignore
					}
				}
				break;
			}
		}
	}


	protected void showSpacesWarning()
	{
		ResourceBundle textsBundle = ResourceBundle.getBundle(
			DEFAULT_BUNDLE_PATH, Locale.getDefault());
		JOptionPane.showMessageDialog(null,
			textsBundle.getString("DjVuFileParser.filenameWarning.message"), //$NON-NLS-1$
			textsBundle.getString("DjVuFileParser.filenameWarning.title"), //$NON-NLS-1$
			JOptionPane.WARNING_MESSAGE);
	}


	public boolean supports(URL fileURL)
	{
		String path = fileURL.getPath();
		String fileName = new File(path).getName();
		return SUPPORTED_MIMETYPE.equalsIgnoreCase(mimetypesFileTypeMap
				.getContentType(fileName));
	}


	public void initialize(Map<String, Object> initPrefs)
	{
		// no information is needed
	}


	@Override
	public String toString()
	{
		return "DjVu parser"; //$NON-NLS-1$
	}
}
