package pl.psnc.dlibra.app.extension.xml;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import pl.psnc.dlibra.app.common.component.metadata.ConversionRules;


/**
 * Conversion rules for XML format. 
 * @version $Id: XMLConversionRules.java,v 1.1 2007-01-09 06:43:19 tparkola Exp $
 */
public class XMLConversionRules
	implements ConversionRules<Map<TestXQuery, Map<String, List<ConversionXQuery>>>>
{

	private Map<TestXQuery, Map<String, List<ConversionXQuery>>> conversionRules;
	
	private static final String CONV_SEPARATOR = ".";

	/**
	 * Initializes conversion rules.
	 * @param tests contains tests that test type of file
	 * @param convs conversion rules for all supported XML file types
	 */
	public XMLConversionRules(Properties tests, Properties convs)
	{
		loadConfiguration(tests, convs);
	}
	
	private void loadConfiguration(Properties tests, Properties convs) 
	{	
		conversionRules = new HashMap<TestXQuery, Map<String,List<ConversionXQuery>>>();
		
		for (Map.Entry e: tests.entrySet()) {
			conversionRules.put(new TestXQuery((String)e.getValue(), (String)e.getKey()), 
				getConversionRules((String)e.getKey(), convs));
		}
	}
	
	private Map<String, List<ConversionXQuery>> getConversionRules(String string, Properties convs)
	{
		Map<String, List<ConversionXQuery>> result = new HashMap<String, List<ConversionXQuery>>();
		
		for (Map.Entry e: convs.entrySet()) {
			if (((String)e.getKey()).startsWith(string + CONV_SEPARATOR)) {
				String attrId = ((String)e.getKey()).substring((string + CONV_SEPARATOR).length());
				List<ConversionXQuery> queries = result.get(attrId);				
				if (queries == null) {
					result.put(attrId, queries = new ArrayList<ConversionXQuery>());
				}
				
				queries.add(new ConversionXQuery((String)e.getValue()));
			}
		}
		
		return result;
	}

	public Map<TestXQuery, Map<String, List<ConversionXQuery>>> getConversionRules()
	{
		return conversionRules;
	}

}
