package pl.psnc.dlibra.app.extension.xml;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Map.Entry;

import net.sf.saxon.javax.xml.xquery.XQConnection;
import net.sf.saxon.javax.xml.xquery.XQException;
import net.sf.saxon.javax.xml.xquery.XQExpression;
import net.sf.saxon.xqj.SaxonXQDataSource;
import pl.psnc.dlibra.app.common.component.metadata.AbstractConverter;
import pl.psnc.dlibra.app.extension.metadata.MetadataFinder;
import pl.psnc.dlibra.app.extension.metadata.MetadataFinderException;
import pl.psnc.dlibra.app.extension.metadata.MetadataFinderException.Operation;


/**
 * Converts metadata in XML format to dLibra metadata set.
 * @version $Id: XMLMetadataConverter.java,v 1.2 2007-02-28 09:11:04 tparkola Exp $
 */
public class XMLMetadataConverter
	extends AbstractConverter<XMLConversionRules>
{	
	private ResourceBundle textsBundle;

	/**
	 * Initializes this converter. The rules will be used in conversion operation.
	 * @param rules
	 */
	public XMLMetadataConverter(XMLConversionRules rules) {
		super(rules);
	}

	public List<Map<String, Map<String, List<String>>>> convert(File file)
	throws MetadataFinderException
	{
		ArrayList<Map<String, Map<String, List<String>>>> metadataSets = 
			new ArrayList<Map<String, Map<String, List<String>>>>();
		
		Map<String, Map<String, List<String>>> metadata = 
			new HashMap<String, Map<String, List<String>>>();
		
		SaxonXQDataSource source = new SaxonXQDataSource();
		XQConnection conn = null;
		XQExpression expr = null;
		try {
			conn = source.getConnection();
			expr = conn.createExpression();
				
			for (Entry<TestXQuery, Map<String, List<ConversionXQuery>>> test: conversionRules.getConversionRules().entrySet()) {
				TestXQuery tq = test.getKey();
				if (tq.executeQuery(file, expr, MetadataFinder.DEFAULT_LANGUAGE).size() < 1) {
					// test failed
					continue;
				}
				
				for (Entry<String, List<ConversionXQuery>> e: test.getValue().entrySet()) {
					List<ConversionXQuery> queries = e.getValue();
					
					for (ConversionXQuery q: queries) {
						Map<String, List<String>> result = q.executeQuery(file, expr, MetadataFinder.DEFAULT_LANGUAGE);
					
						
						for (Entry<String, List<String>> entry: result.entrySet()) {						
							Map<String, List<String>> rdfToValues = metadata.get(entry.getKey());
							if (rdfToValues == null) {
								metadata.put(entry.getKey(), 
									rdfToValues = new HashMap<String, List<String>>());
							}
							List<String> currentValues = rdfToValues.get(e.getKey());
							if (currentValues == null) {
								rdfToValues.put(e.getKey(), 
									currentValues = new ArrayList<String>());
							}
							currentValues.addAll(entry.getValue());
						}					
					}								
				}
			}
		}
		catch (XQException e1) {			
			throw new MetadataFinderException(Operation.Load, 
				textsBundle.getString("LoadMetadata.invalidFormat.reason"));
		}
		finally {
			if (expr != null) {
				expr.close();
			}
			try {
				if (conn != null) {
					conn.close(); 
				}
			}
			catch (XQException e) {
				// do nothing
			}
		}
		
		if (metadata.size() > 0) {
			metadataSets.add(metadata);
		}
		return metadataSets;
	}
	
	/**
	 * @param textsBundle bundle for exception messages
	 */
	public void setResourceBundle(ResourceBundle textsBundle)
	{
		this.textsBundle = textsBundle;
	}
}
