package pl.psnc.dlibra.app.plugin.xml;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.List;
import java.util.Map;

import junit.framework.JUnit4TestAdapter;
import net.sf.saxon.javax.xml.xquery.XQException;
import net.sf.saxon.javax.xml.xquery.XQExpression;
import net.sf.saxon.xqj.SaxonXQDataSource;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import pl.psnc.dlibra.app.extension.xml.ConversionXQuery;
import pl.psnc.dlibra.app.extension.xml.TestXQuery;


public class XQueryTest
{
	private File metadata;

	@Before
	public void createTestFiles() throws IOException {
		metadata = File.createTempFile("xmlfile", "test");
		
		BufferedWriter bw = new BufferedWriter(new FileWriter(metadata));
		bw.write("<test><abc>the value</abc></test>");
		bw.close();		
	}
	
	@After
	public void deleteTestFiles() {
		metadata.delete();
	}
	
	@Test
	public void extecuteConversionXQuery() throws XQException {
		String string = "for $x in fn:doc({document})/test/abc return $x";
		ConversionXQuery xq = new ConversionXQuery(string);
		
		XQExpression expr = new SaxonXQDataSource().getConnection().createExpression();
		Map<String, List<String>> result = xq.executeQuery(metadata, expr, "myDefaultLang");
		
		assertNotNull(result.get("myDefaultLang"));
		assertEquals("the value", result.get("myDefaultLang").get(0));
	}
	
	@Test
	public void extecuteTestXQuery() throws XQException {
		String string = "for $x in fn:doc({document})/test/abc return $x";
		TestXQuery xq = new TestXQuery(string, "master");
		
		XQExpression expr = new SaxonXQDataSource().getConnection().createExpression();
		Map<String, List<String>> result = xq.executeQuery(metadata, expr, "myDefaultLang");
		
		assertNotNull(result.get("myDefaultLang"));
		assertEquals("the value", result.get("myDefaultLang").get(0));
		assertEquals("master", xq.getXMLFileType());
	}
	
	public static junit.framework.Test suite() {
		return new JUnit4TestAdapter(XQueryTest.class);
	}
}
