/**
 * $Id: LeafsListTool.java,v 1.2 2010/02/09 13:39:00 tparkola Exp $
 *
 * Copyright (c) 2002 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.app.extension.tl.leafslist;

import java.awt.Window;
import java.awt.event.ActionEvent;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Locale;
import java.util.Properties;
import java.util.ResourceBundle;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;

import pl.psnc.dlibra.app.common.IOUtilities;
import pl.psnc.dlibra.app.common.component.configuration.DefaultConfigurationDialog;
import pl.psnc.dlibra.app.common.component.extension.AbstractConfigurableExtension;
import pl.psnc.dlibra.app.extension.tool.SelectionController;
import pl.psnc.dlibra.app.extension.tool.Tool;
import pl.psnc.dlibra.common.Info;
import pl.psnc.dlibra.common.OutputFilter;
import pl.psnc.dlibra.metadata.AbstractPublicationInfo;
import pl.psnc.dlibra.metadata.MetadataServer;
import pl.psnc.dlibra.metadata.Publication;
import pl.psnc.dlibra.metadata.PublicationFilter;
import pl.psnc.dlibra.metadata.PublicationId;
import pl.psnc.dlibra.mgmt.AbstractServiceResolver;

public class LeafsListTool
	extends AbstractConfigurableExtension
	implements Tool
{

	public static final String VELOCITY_TEMPLATE = "index.vm";


	public LeafsListTool()
	{
		DEFAULT_BUNDLE_PATH = "pl/psnc/dlibra/app/extension/tl/leafslist/GUI";
		DEFAULT_CONFIG_FILES_PACKAGE = "/pl/psnc/dlibra/app/extension/tl/leafslist/template/";
		reloadResourceBundle(DEFAULT_BUNDLE_PATH);
	}


	@Override
	protected void reloadResourceBundle(String bundlePath)
	{
		textsBundle = ResourceBundle.getBundle(bundlePath, Locale.getDefault());
	}


	public String getName()
	{
		if (!Locale.getDefault().equals(textsBundle.getLocale())) {
			reloadResourceBundle(DEFAULT_BUNDLE_PATH);
		}
		return textsBundle.getString("Name");
	}


	public void perform(Window parentWindow,
			SelectionController externalController, Object serverInterface)
	{
		if (!Locale.getDefault().equals(textsBundle.getLocale())) {
			reloadResourceBundle(DEFAULT_BUNDLE_PATH);
		}
		File template = getFile(VELOCITY_TEMPLATE);

		String id = JOptionPane.showInputDialog(parentWindow,
			textsBundle.getString("providePubId.text"),
			textsBundle.getString("generatePubLeafList.title"),
			JOptionPane.PLAIN_MESSAGE);

		if (id == null) {
			// user cancelled operation
			return;
		}

		Long pubId = null;

		try {
			pubId = Long.parseLong(id);
		}
		catch (NumberFormatException e) {
			JOptionPane.showMessageDialog(parentWindow,
				textsBundle.getString("wrongPubId.msg"),
				textsBundle.getString("wrongPubId.title"),
				JOptionPane.ERROR_MESSAGE);
			return;
		}

		AbstractServiceResolver usr = (AbstractServiceResolver) serverInterface;

		Collection<Info> listOfPubs = null;
		try {
			listOfPubs = ((MetadataServer) usr.getService(
				MetadataServer.SERVICE_TYPE, null))
					.getPublicationManager()
					.getObjects(
						new PublicationFilter(null, new PublicationId(pubId))
								.setGroupStatus(Publication.PUB_GROUP_LEAF)
								.setRecursive(true),
						new OutputFilter(AbstractPublicationInfo.class))
					.getResultInfos();
			List<AbstractPublicationInfo> publications = new ArrayList<AbstractPublicationInfo>();

			for (Info i : listOfPubs) {
				if (i instanceof AbstractPublicationInfo) {
					publications.add((AbstractPublicationInfo) i);
				}
			}

			Collections.sort(publications,
				new Comparator<AbstractPublicationInfo>() {

					public int compare(AbstractPublicationInfo o1,
							AbstractPublicationInfo o2)
					{
						return Collator.getInstance().compare(o1.getLabel(),
							o2.getLabel());
					}
				});

			JFileChooser saver = new JFileChooser();

			if (saver.showSaveDialog(parentWindow) == JFileChooser.APPROVE_OPTION) {
				File selected = saver.getSelectedFile();

				VelocityEngine ve = new VelocityEngine();
				Properties props = new Properties();
				props.put("resource.loader", "file");
				props.put("file.resource.loader.class",
					"org.apache.velocity.runtime.resource.loader.FileResourceLoader");
				props.put("file.resource.loader.path", ""); // !!!
				props.put("file.resource.loader.cache", "true");
				props.put("runtime.log.logsystem.class",
					"org.apache.velocity.runtime.log.NullLogSystem");
				ve.init(props);

				PrintWriter print = new PrintWriter(new BufferedWriter(
						new FileWriter(selected)));

				VelocityContext context = new VelocityContext();

				context.put("pubs", publications);

				StringWriter sw = new StringWriter();
				ve.evaluate(context, sw, "", new InputStreamReader(
						new FileInputStream(template), "utf-8"));
				print.println(sw.toString());
				print.close();

				JOptionPane.showMessageDialog(parentWindow,
					textsBundle.getString("success.msg"),
					textsBundle.getString("success.title"),
					JOptionPane.INFORMATION_MESSAGE);
			}
		}
		catch (Exception e) {
			JOptionPane
					.showMessageDialog(parentWindow,
						textsBundle.getString("error.msg"),
						textsBundle.getString("error.title"),
						JOptionPane.ERROR_MESSAGE);
		}
	}


	@Override
	public void applyActionPerformed(ActionEvent e)
	{
		super.applyActionPerformed(e);

		List<String> filesNames = new ArrayList<String>();
		filesNames.add(VELOCITY_TEMPLATE);

		retrieveConfigurationFiles(e, filesNames);

		getFile(VELOCITY_TEMPLATE);
	}


	private File getFile(String fileName)
	{
		File result = new File(localConfigDir, fileName);
		Object useExtFlag = configuration
				.get(DefaultConfigurationDialog.USE_EXTERNAL_CONFIG_DIR);
		if (useExtFlag != null && ((Boolean) useExtFlag).booleanValue()) {
			if (!result.exists()) {
				try {
					IOUtilities.copyStream(
						this.getClass().getResourceAsStream(
							DEFAULT_CONFIG_FILES_PACKAGE + fileName),
						new FileOutputStream(result));
				}
				catch (IOException e1) {
					// could not load configuration
				}
			}

			return result;
		}

		try {
			IOUtilities.copyStream(
				this.getClass().getResourceAsStream(
					DEFAULT_CONFIG_FILES_PACKAGE + fileName),
				new FileOutputStream(result));
		}
		catch (IOException e1) {
			// could not load configuration
		}

		return result;
	}


	@Override
	public String toString()
	{
		return getName();
	}

	
}