/*
 * Copyright (c) 2006 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.content.extraction;

import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.util.List;

import org.apache.log4j.Logger;
import org.xml.sax.InputSource;
import org.yufeng.jchmlib.ChmManager;
import org.yufeng.jchmlib.FileEntry;

/**
 * Can be used to extract text from RTF files.
 * @version $Id: CHMTextualContentExtractor.java,v 1.2 2007/11/23 08:51:03 helin Exp $
 * @todo Character Encoding is not very good :|
 */
public class CHMTextualContentExtractor
	extends HTMLTextualContentExtractor
{

	/**
	 * Logger for this class
	 */
	private static final Logger logger = Logger
			.getLogger(CHMTextualContentExtractor.class);

	private static final String HTML = ".html";

	private static final String HTM = ".htm";

	private static final String APPLICATION_CHM = "application/chm";


	public String getTextualContent(File file, String mimeType, String version,
			Integer pageNumber, File mainFile)
		throws IOException
	{
		StringBuffer bodyText = new StringBuffer();
		ChmManager chmManager = new ChmManager(file.getAbsolutePath());
		List files = chmManager.enumerateFiles();
		long totalSize = 0;
		long totalFiles = 0;
		long htmlFiles = 0;
		long brokenHtmlFiles = 0;
		try {
			for (Object fileO : files) {
				FileEntry fileE = (FileEntry) fileO;
				totalFiles++;
				String entryName = fileE.entryName;
				if (entryName.endsWith(HTM) || entryName.endsWith(HTML)) {
					htmlFiles++;
					String fileContent = null;
					try {
						fileContent = chmManager.retrieveFile(fileE);
					}
					catch (Exception e) {
						logger.error("Error when loading entry,"
								+ " skipping. File: " + file + ", entry: "
								+ entryName + ". [" + e.getClass().getName()
								+ "]");
						logger.debug(e);
						brokenHtmlFiles++;
						continue;
					}
					if (fileContent != null) {
						//HTML parser should go here
						totalSize += fileContent.length();
						InputSource inputSource = new InputSource(
								new StringReader(fileContent));
						try {
							fileContent = parseInputSource(inputSource);

							StringBuffer footer = new StringBuffer();
							footer.append("\n");
							footer
									.append(TextualContentExtractor.CONTENT_SEPARATOR_PREFIX);
							footer.append(" ");
							footer
									.append(TextualContentExtractor.CONTENT_SEPARATOR_FILE);
							footer.append(version);
							footer.append(" ");
							footer
									.append(TextualContentExtractor.CONTENT_SEPARATOR_PAGE);
							footer.append(totalFiles);
							footer.append(" ");
							footer
									.append(TextualContentExtractor.CONTENT_SEPARATOR_SUFFIX);

							fileContent += footer.toString();
						}
						catch (Exception e) {
							logger.error("Error when parsing HTML,"
									+ " skipping file. File: " + file
									+ ", entry: " + entryName + ". ["
									+ e.getClass().getName() + "]");
							logger.debug(e);
							brokenHtmlFiles++;
							continue;
						}
						bodyText.append(fileContent);
						bodyText.append(' ');
					}
				}
			}
		}
		catch (OutOfMemoryError e) {
			logger.error("Error occured - curently extracted " + totalSize
					+ " bytes of pure text. Skipping the rest", e);
		}
		logger.info("Extracted " + htmlFiles + " html files of total "
				+ totalFiles + " chm file entries. " + brokenHtmlFiles
				+ " of extracted html files was broken."
				+ " Overall text length (in bytes): " + totalSize);
		return bodyText.toString();
	}


	@Override
	public boolean isAbleToExtract(String path, String mimeType)
	{
		return APPLICATION_CHM.equals(mimeType);
	}

}
