/*
 * Copyright (c) 2005 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.content.extraction;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import org.apache.html.dom.HTMLHtmlElementImpl;
import org.apache.html.dom.HTMLMetaElementImpl;
import org.apache.log4j.Logger;
import org.cyberneko.html.parsers.DOMParser;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Class for extracting content of HTML files.
 * @version $Id: HTMLTextualContentExtractor.java,v 1.2 2007/11/23 08:51:03 helin Exp $
 */
public class HTMLTextualContentExtractor
	implements TextualContentExtractor
{

	/**
	 * Logger for this class
	 */
	private static final Logger logger = Logger
			.getLogger(HTMLTextualContentExtractor.class);

	private static final String TEXT_HTML = "text/html";

	private static final String META_TAG_NAME_ATTRIBUTE = "name";


	/**
	 * Parses given input source and removes HTML tags.
	 * @param inputSource Input source with HTML
	 * @return String eith HTML content
	 * @throws SAXException
	 * @throws IOException
	 */
	protected String parseInputSource(InputSource inputSource)
		throws SAXException, IOException
	{
		DOMParser parser = new DOMParser();
		parser.parse(inputSource);
		String text = getText(parser.getDocument());
		return text;
	}


	private String getText(Node node)
	{
		String result = "";
		if (node instanceof HTMLHtmlElementImpl) {
			result = node.getTextContent();
		}
		if (node instanceof HTMLMetaElementImpl) {
			NamedNodeMap atts = node.getAttributes();
			Node name = atts.getNamedItem(META_TAG_NAME_ATTRIBUTE);
			if (name != null) {
				int length = atts.getLength();
				for (int i = 0; i < length; i++) {
					Node n = atts.item(i);
					if (!n.getNodeName().equalsIgnoreCase(
						META_TAG_NAME_ATTRIBUTE)) {
						result = result + " " + n.getNodeValue();
					}
				}
			}
		}
		Node child = node.getFirstChild();
		while (child != null) {
			result = getText(child) + " " + result;
			child = child.getNextSibling();
		}
		return result;
	}


	public String getTextualContent(File file, String mimeType, String version,
			Integer pageNumber, File mainFile)
		throws IOException
	{
		String text;

		try {
			InputSource inputSource = new InputSource(new FileInputStream(file));
			text = parseInputSource(inputSource);

			StringBuffer footer = new StringBuffer();
			footer.append("\n");
			footer.append(TextualContentExtractor.CONTENT_SEPARATOR_PREFIX);
			footer.append(" ");
			footer.append(TextualContentExtractor.CONTENT_SEPARATOR_FILE);
			footer.append(version);
			footer.append(" ");
			footer.append(TextualContentExtractor.CONTENT_SEPARATOR_PAGE);
			if (pageNumber == null)
				pageNumber = Integer.valueOf(0);
			footer.append(pageNumber);
			footer.append(" ");
			footer.append(TextualContentExtractor.CONTENT_SEPARATOR_SUFFIX);

			text += footer.toString();

			return text;
		}
		catch (Exception e) {
			logger.error("HTML Parser error: " + e.getMessage());
			throw new IOException(e.getMessage());
		}
	}


	public boolean isAbleToExtract(String path, String mimeType)
	{
		return TEXT_HTML.equals(mimeType);
	}

}
