/*
 * Copyright (c) 2006 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.content.extraction;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;

import javax.swing.text.DefaultStyledDocument;
import javax.swing.text.rtf.RTFEditorKit;

import org.apache.log4j.Logger;

/**
 * Can be used to extract text from RTF files.
 * @version $Id: RTFTextualContentExtractor.java,v 1.2 2007/11/23 08:51:03 helin Exp $
 * @todo Character Encoding is not very good :|
 */
public class RTFTextualContentExtractor
	implements TextualContentExtractor
{

	private static final String APPLICATION_RTF = "application/rtf";

	/**
	 * Logger for this class
	 */
	private static final Logger logger = Logger
			.getLogger(RTFTextualContentExtractor.class);


	public String getTextualContent(File file, String mimeType, String version,
			Integer pageNumber, File mainFile)
		throws FileNotFoundException
	{
		String bodyText = null;

		DefaultStyledDocument styledDoc = new DefaultStyledDocument();
		try {
			InputStream inputStream = new FileInputStream(file);
			new RTFEditorKit().read(inputStream, styledDoc, 0);
			bodyText = styledDoc.getText(0, styledDoc.getLength());
			inputStream.close();
		}
		catch (Exception e) {
			logger.warn("Cannot extract text from a RTF document", e);
		}

		if (bodyText != null) {
			try {
				bodyText = new String(bodyText.getBytes("ISO-8859-1"),
						"Windows-1250");

				StringBuffer footer = new StringBuffer();
				footer.append("\n");
				footer.append(TextualContentExtractor.CONTENT_SEPARATOR_PREFIX);
				footer.append(" ");
				footer.append(TextualContentExtractor.CONTENT_SEPARATOR_FILE);
				footer.append(version);
				footer.append(" ");
				footer.append(TextualContentExtractor.CONTENT_SEPARATOR_PAGE);
				if (pageNumber == null)
					pageNumber = Integer.valueOf(0);
				footer.append(pageNumber);
				footer.append(" ");
				footer.append(TextualContentExtractor.CONTENT_SEPARATOR_SUFFIX);

				bodyText += footer.toString();
			}
			catch (UnsupportedEncodingException e) {
				logger.error("This exception should not occur!", e);
			}
			return bodyText;
		}
		return null;
	}


	public boolean isAbleToExtract(String path, String mimeType)
	{
		return APPLICATION_RTF.equals(mimeType);
	}

}
