package pl.psnc.dlibra.content.extraction;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;

import org.apache.log4j.Logger;

import com.lizardtech.djvu.IdentifierConverter;

public class DjvuIdentifierConverterImpl
	implements IdentifierConverter
{

	private static Logger logger = Logger
			.getLogger(DjvuIdentifierConverterImpl.class.getName());

	private static final String DEFAULT_DJBZ_FILE = "default.djbz";

	private static final String DEFAULT_SHARED_ANNO_FILE = "default_shared_anno.iff";

	private static final String DJBZ_EXTENSION = ".djbz";

	private static final String SHARED_ANNO_EXTENSION = ".iff";


	/**
	 * Converts arg1 to the URL appropriate for the given publication and
	 * edition
	 * 
	 * @param arg0
	 *            helper url containing publication identifier
	 * @param arg1
	 *            file name of the file to convert; maybe a djvu name or version
	 *            name
	 */
	public URL convertId(URL arg0, String arg1)
	{
		String fileName;

		if (arg1.endsWith(DJBZ_EXTENSION))
			fileName = DEFAULT_DJBZ_FILE;
		else if (arg1.endsWith(SHARED_ANNO_EXTENSION))
			fileName = DEFAULT_SHARED_ANNO_FILE;
		else
			return arg0;

		try {
			File f = new File(System.getProperty("java.io.tmpdir"), fileName);
			if (!f.exists())
				copyFile(getFile(fileName), f);
			f.deleteOnExit();
			return f.toURI().toURL();
		}
		catch (MalformedURLException e) {
			logger.warn("Converting djvu identifier failed...", e);
		}
		catch (IOException e) {
			logger.warn("Converting djvu identifier failed...", e);
		}

		return arg0;
	}


	private InputStream getFile(String fileName)
		throws FileNotFoundException
	{
		ClassLoader classLoader = this.getClass().getClassLoader();
		if (classLoader == null)
			throw new FileNotFoundException(fileName);
		return classLoader.getResourceAsStream(fileName);
	}


	private void copyFile(InputStream is, File dest)
		throws IOException
	{
		if (is == null)
			return;

		OutputStream out = null;
		try {
			out = new FileOutputStream(dest);
			byte[] buf = new byte[1024];
			int len;
			while ((len = is.read(buf)) > 0) {
				out.write(buf, 0, len);
			}
		}
		finally {
			try {
				is.close();
			}
			finally {
				if (out != null)
					out.close();
			}
		}

	}
}
