/*
 * Copyright (c) 2005 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.content.extraction;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import lius.config.LiusConfig;
import lius.index.Indexer;
import lius.index.IndexerFactory;

/**
 * Class for extracting content of PDF files.
 * @version $Id: LIUSTextualContentExtractor.java,v 1.2 2007/11/23 08:51:02 helin Exp $
 */
public class LIUSTextualContentExtractor
	implements TextualContentExtractor
{

	private static LiusConfig liusConfig;


	public String getTextualContent(File file, String mimeType, String version,
			Integer pageNumber, File mainFile)
		throws FileNotFoundException, IOException
	{
		Indexer indexer = IndexerFactory.getIndexer(new FileInputStream(file),
			mimeType, liusConfig);
		if (indexer != null) {
			StringBuffer content = new StringBuffer(indexer.getContent());
			content.append("\n");
			content.append(TextualContentExtractor.CONTENT_SEPARATOR_PREFIX);
			content.append(" ");
			content.append(TextualContentExtractor.CONTENT_SEPARATOR_FILE);
			content.append(version);
			content.append(" ");
			content.append(TextualContentExtractor.CONTENT_SEPARATOR_PAGE);
			content.append(pageNumber != null ? String.valueOf(pageNumber) : 0);
			content.append(" ");
			content.append(TextualContentExtractor.CONTENT_SEPARATOR_SUFFIX);

			return content.toString();
		}
		return null;
	}


	public boolean isAbleToExtract(String path, String mimeType)
	{
		boolean isAble;

		try {
			isAble = (IndexerFactory.getIndexer(new FileInputStream(new File(
					path)), mimeType, liusConfig) != null);
		}
		catch (FileNotFoundException e) {
			isAble = false;
		}

		return isAble;
	}


	/**
	 * @param liusConfig the liusConfig to set
	 */
	public static final void setLiusConfig(LiusConfig liusConfig)
	{
		LIUSTextualContentExtractor.liusConfig = liusConfig;
	}

}
