/*
 * Copyright (c) 2005 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.content.extraction;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.List;

import org.apache.log4j.Logger;
import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDDocumentCatalog;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.pdmodel.interactive.annotation.PDAnnotation;
import org.apache.pdfbox.pdmodel.interactive.documentnavigation.outline.PDDocumentOutline;
import org.apache.pdfbox.pdmodel.interactive.documentnavigation.outline.PDOutlineItem;
import org.apache.pdfbox.pdmodel.interactive.documentnavigation.outline.PDOutlineNode;

/**
 * Class for extracting content of PDF files.
 * @version $Id: PDFTextualContentExtractor.java,v 1.4 2009/06/22 11:56:42 helin Exp $
 */
public class PDFTextualContentExtractor
	implements TextualContentExtractor
{

	private static final String APPLICATION_PDF = "application/pdf";

	/**
	 * Logger for this class
	 */
	private static final Logger logger = Logger
			.getLogger(PDFTextualContentExtractor.class);


	public String getTextualContent(File file, String mimeType, String version,
			Integer pageNumber, File mainFile)
		throws FileNotFoundException, IOException
	{
		PDDocument pdfDocument = PDDocument.load(file);
		if (pdfDocument.isEncrypted()) {
			try {
				pdfDocument.decrypt("");
			}
			catch (Throwable e) {
				logger.warn("Cannot index documents encrypted with "
						+ "non empty password (" + file + "). "
						+ e.getClass().getName().toString() + ": "
						+ e.getMessage());
				pdfDocument.close();
				return null;
			}
		}
		String content = null;
		try {
			//text
			DlibraPDFTextStripper stripper = new DlibraPDFTextStripper(version);
			content = stripper.getText(pdfDocument);
			//bookmarks
			PDDocumentOutline outline = pdfDocument.getDocumentCatalog()
					.getDocumentOutline();
			if (outline != null) {
				content += getBookmarksTitles(outline);
			}
			// annotations
			content += " " + getAnnotations(pdfDocument.getDocumentCatalog());
		}
		catch (IOException e) {
			logger.warn("IO exception while indexing PDF " + file
					+ ". Message: " + e.getMessage());
			pdfDocument.close();
			return null;
		}
		pdfDocument.close();
		return content;
	}


	private String getBookmarksTitles(PDOutlineNode bookmark)
		throws IOException
	{
		PDOutlineItem current = bookmark.getFirstChild();
		StringBuffer bookmarks = new StringBuffer();
		while (current != null) {
			bookmarks.append(" ");
			bookmarks.append(current.getTitle());
			bookmarks.append(getBookmarksTitles(current));
			current = current.getNextSibling();
		}
		return bookmarks.toString();
	}
	
	private String getAnnotations(PDDocumentCatalog documentCatalog)
	{
		StringBuffer buffer = new StringBuffer();
		List<PDPage> pages = documentCatalog.getAllPages();
		for (PDPage page : pages) {
			try {
				List<PDAnnotation> annotations = page.getAnnotations();
				for (PDAnnotation annotation : annotations) {
					String contents = annotation.getContents();
					if (contents != null)
						buffer.append(contents);
				}
			}
			catch (Exception e) {
				logger.warn("Can't get annotations from page!");
			}
		}
		return buffer.toString();
	}


	public boolean isAbleToExtract(String path, String mimeType)
	{
		return APPLICATION_PDF.equals(mimeType);
	}

}
