package pl.psnc.dlibra.content.extraction;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;

import org.apache.log4j.Logger;
import org.xml.sax.SAXException;

import pl.psnc.dlibra.app.extension.dsurl.URLPublicationBuilder;
import pl.psnc.dlibra.plugin.TextualContentExtractorFactory;
import pl.psnc.util.IOUtils;

/**
 * Class for extracting content linked to by URL publications.
 */
public class URLTextualContentExtractor
	implements TextualContentExtractor
{

	private static final String TEXT_XML = "text/xml";

	/**
	 * Logger for this class
	 */
	private static final Logger logger = Logger
			.getLogger(TextualContentExtractor.class);


	public String getTextualContent(File file, String mimeType, String version,
			Integer pageNumber, File mainFile)
		throws IOException
	{
		StringBuilder sb = new StringBuilder();
		try {
			// parse xml file, collect links that lead to the text to index
			List<URL> indexLinks = new ArrayList<URL>();
			URLPublicationBuilder.loadURLPublicationFile(file, indexLinks);

			for (URL link : indexLinks) {
				sb.append(getContent(link, version, pageNumber, mainFile))
						.append("\n");
			}
		}
		catch (ParserConfigurationException e) {
			logger.warn("Error parsing file " + file.getAbsolutePath(), e);
			return null;
		}
		catch (SAXException e) {
			logger.warn("Error parsing file " + file.getAbsolutePath(), e);
			return null;
		}

		return sb.toString();
	}


	private File storeRemoteContent(InputStream is, String link)
		throws IOException
	{
		String tmpPrefix = link.replaceAll("[^a-zA-Z0-9]", "");
		tmpPrefix = tmpPrefix.substring(Math.max(0, tmpPrefix.length() - 20));
		File f = File.createTempFile(tmpPrefix, null);

		try {
			FileOutputStream out = new FileOutputStream(f);
			try {
				IOUtils.copyStream(is, out);
			}
			finally {
				out.close();
			}
		}
		catch (IOException e) {
			f.delete();
			throw e;
		}
		finally {
			is.close();
		}

		return f;
	}


	private String getContent(URL link, String version,
			Integer pageNumber, File mainFile)
		throws IOException
	{
		URLConnection connection = link.openConnection();
		String contentType = connection.getContentType();
		int parametersPos = contentType.indexOf(';');
		if (parametersPos >= 0) {
			contentType = contentType.substring(0, parametersPos);
		}

		File f = storeRemoteContent(
			new BufferedInputStream(connection.getInputStream()),
			link.getFile());

		try {
			TextualContentExtractor extractor = TextualContentExtractorFactory
					.getExtractor(f.getAbsolutePath(), contentType);
			// TODO check whether version, pageNumber and mainFile passed here are ok
			return (extractor == null) ? null : extractor.getTextualContent(f,
				contentType, version, pageNumber, mainFile);
		}
		finally {
			// clean
			f.delete();
		}
	}


	public boolean isAbleToExtract(String path, String mimeType)
	{
		if (!TEXT_XML.equals(mimeType))
			return false;

		// parse xml
		try {
			return (URLPublicationBuilder.loadURLPublicationFile(
				new File(path), null) != null);
		}
		catch (Exception e) {
			return false;
		}
	}
}
