/**
 * $Id: CustomImportStyleSheet.java,v 1.4 2006-05-29 13:36:09 mwerla Exp $
 *
 * Copyright (c) 2002 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package javax.swing.text.html;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.net.URL;

import pl.psnc.util.crypto.CryptoUtils;

/**
 * This class is to support encrypted retrieval of CSS styles.
 * @version $Id: CustomImportStyleSheet.java,v 1.4 2006-05-29 13:36:09 mwerla Exp $
 */
public class CustomImportStyleSheet
	extends StyleSheet
{

	/**
	 * a secretKey needed to get encrypted content
	 */
	private CryptoUtils utils;


	/**
	 * Constructor
	 * @param utils to get encrypted content
	 */
	public CustomImportStyleSheet(CryptoUtils utils)
	{
		this.utils = utils;
	}


	/**
	 * Rertievs style file from a given url using appriopriate encryption.
	 * Uses reflection to use javax.swing.text.html.StyleSheet.CssParser.
	 * @param url of the style file.
	 */
	@Override
	public void importStyleSheet(URL url)
	{
		try {
			InputStream encInputStream = utils
					.getDecipheredInputStream(url.openStream());
			String encoding = "UTF-8";
			BufferedReader r = new BufferedReader(new InputStreamReader(
					encInputStream, encoding));

			Class ss = StyleSheet.class;
			Class cssParserClass = java.lang.ClassLoader.getSystemClassLoader()
					.loadClass("javax.swing.text.html.StyleSheet$CssParser");
			Class classes[] = ss.getDeclaredClasses();
			for (int i = 0; i < classes.length; i++) {
				Class clazz = classes[i];
				if (clazz.equals(cssParserClass)) {
					Constructor c = clazz
							.getDeclaredConstructor(new Class[] { StyleSheet.class});
					c.setAccessible(true);
					Object o = c.newInstance(new Object[] { this});
					Method parse = clazz.getDeclaredMethod("parse",
						new Class[] { URL.class, Reader.class, boolean.class,
								boolean.class});
					parse.setAccessible(true);
					parse.invoke(o, new Object[] { url, r, Boolean.FALSE,
							Boolean.TRUE});
					break;
				}
			}
			r.close();
			encInputStream.close();
		}
		catch (Throwable e) {
			//simply: we won't have styles...
			e.printStackTrace();
		}
	}
}
