/*
 * Copyright (c) 2000 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */

package javax.swing.text.html;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringWriter;

import javax.swing.JEditorPane;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.Element;
import javax.swing.text.StyleConstants;
import javax.swing.text.View;
import javax.swing.text.ViewFactory;

import pl.psnc.util.crypto.CryptoUtils;

/**
 * An HTMLEditorKit that supports encrypted content.
 */
public class ViewerHTMLEditorKit
	extends HTMLEditorKit
{

	private static final ViewerHTMLFactory viewerFactory = new ViewerHTMLFactory();

	private static CryptoUtils utils;

	private static String sessionId;

	private JEditorPane editorPane;

	private Parser defaultParser;


	/**
	 * Sets the CryptoUtils for deciphering.
	 * @param utils
	 */
	public static void setCryptoUtils(CryptoUtils utils)
	{
		ViewerHTMLEditorKit.utils = utils;
	}

	/**
	 * Sets base publiaction url.
	 * @param url
	 */
	public static void setBaseUrl(String url)
	{
		ViewerHTMLFactory.setBaseUrl(url);
	}


	/**
	 * Sets user session id.
	 * @param sessionId
	 */
	public static void setSessionId(String sessionId)
	{
		ViewerHTMLEditorKit.sessionId = sessionId;
	}


	/**
	 * This implementation works the same as the original,
	 * only the parser instance is not shared between instances
	 * of this class (parser is stored in non-static field).
	 * This is a magical fix for DLI-2828.
	 */
	@Override
	protected Parser getParser()
	{
		if (defaultParser == null) {
		    try {
	                Class< ? > c = Class.forName("javax.swing.text.html.parser.ParserDelegator");
	                defaultParser = (Parser) c.newInstance();
		    } catch (Throwable e) {
		    	e.printStackTrace();
		    }
		}
		return defaultParser;
	}


	@Override
	public ViewFactory getViewFactory()
	{
		return viewerFactory;
	}


	@Override
	public void install(JEditorPane editorPane)
	{
		this.editorPane = editorPane;
		super.install(editorPane);
	}


	@Override
	public void read(Reader in, Document doc, int pos)
		throws IOException, BadLocationException
	{
		if (in instanceof InputStreamReader && utils.isKeySet()) {
			StringWriter stringWriter = new StringWriter();
			char[] buff = new char[4096];
			int charsRead = 0;

			while ((charsRead = in.read(buff)) > 0) {
				stringWriter.write(buff, 0, charsRead);
			}

			in.close();

			String charset = (String) editorPane.getClientProperty("charset");

			ByteArrayInputStream byteArrayInputStream;

			if (charset != null) {
				byteArrayInputStream = new ByteArrayInputStream(stringWriter
						.toString().getBytes(
							(String) editorPane.getClientProperty("charset")));
			}
			else {
				byteArrayInputStream = new ByteArrayInputStream(stringWriter
						.toString().getBytes());
			}

			if (charset != null) {
				super.read(new InputStreamReader(utils.getDecipheredInputStream(
					byteArrayInputStream), (String) editorPane
						.getClientProperty("charset")), doc, pos);
			}
			else {
				super.read(new InputStreamReader(utils.getDecipheredInputStream(
					byteArrayInputStream)), doc, pos);
			}
		}
		else {
			super.read(in, doc, pos);
		}
	}

	/**
	 *
	 */
	public static class ViewerHTMLFactory
		extends HTMLEditorKit.HTMLFactory
	{

		private static String baseUrl;


		@Override
		public View create(Element elem)
		{
			// Override the default behaviour for the IMG tag
			Object o = elem.getAttributes().getAttribute(
				StyleConstants.NameAttribute);

			if (o instanceof HTML.Tag) {
				HTML.Tag kind = (HTML.Tag) o;

				if (kind == HTML.Tag.IMG) {
					return new ViewerImageView(elem, sessionId, baseUrl, utils);
				}
			}

			return super.create(elem);
		}


		/**
		 * @param baseUrl The baseUrl to set.
		 */
		public static void setBaseUrl(String baseUrl)
		{
			ViewerHTMLFactory.baseUrl = baseUrl;
		}
	}
}
