/*
 * Copyright (c) 2000 Poznan Supercomputing and Networking Center 10
 * Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland All rights reserved.
 *
 * This software is the confidential and proprietary information of Poznan
 * Supercomputing and Networking Center ("Confidential Information"). You shall
 * not disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into with
 * PSNC.
 */

package javax.swing.text.html;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Toolkit;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.ImageObserver;
import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Dictionary;

import javax.crypto.SecretKey;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JEditorPane;
import javax.swing.event.DocumentEvent;
import javax.swing.text.AbstractDocument;
import javax.swing.text.AttributeSet;
import javax.swing.text.Document;
import javax.swing.text.Element;
import javax.swing.text.JTextComponent;
import javax.swing.text.MutableAttributeSet;
import javax.swing.text.Position;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyledDocument;
import javax.swing.text.View;
import javax.swing.text.ViewFactory;

import pl.psnc.util.crypto.CryptoUtils;

/**
 * @version $Id: ViewerImageView.java,v 1.9 2006-07-31 12:58:55 mwerla Exp $
 */
class ViewerImageView
	extends View
	implements ImageObserver, MouseListener, MouseMotionListener
{

	/**
	 * Static properties for incremental drawing. Swiped from Component.java
	 *
	 * @see #imageUpdate
	 */
	private static boolean sIsInc = true;

	private static int sIncRate = 100;

	private static Icon sPendingImageIcon;

	private static Icon sMissingImageIcon;

	private static final String PENDING_IMAGE_SRC = "icons/image-delayed.gif";

	private static final String MISSING_IMAGE_SRC = "icons/image-failed.gif";

	/** @todo Move this someplace public */
	static final String IMAGE_CACHE_PROPERTY = "imageCache";

	private static final int DEFAULT_WIDTH = 32;

	private static final int DEFAULT_HEIGHT = 32;

	/** @todo possibly move into stylesheet? */
	private static final int DEFAULT_BORDER = 2;

	private AttributeSet attributes;

	private Component fComponent;

	private Container fContainer;

	private Element fElement;

	private Image fImage;

	/** Base of drag while growing image. */
	private Point fGrowBase;

	private Rectangle fBounds = new Rectangle();

	private String sessionId;

	private boolean isLink;

	/**
	 * Set to true, while the receiver is locked, to indicate that the receiver
	 * is loading the image. This is used in imageUpdate.
	 */
	private boolean loading;

	private int border;

	private int fHeight;

	private int fWidth;

	private int xSpace;

	private int ySpace;

	private String baseUrl;

	private CryptoUtils cryptoUtils;

	/**
	 * Creates a new view that represents an IMG element.
	 *
	 * @param elem
	 *            The element to create a view for.
	 * @param sessionId Session id
	 * @param baseUrl base URL.
	 * @param key cipher key
	 */
	public ViewerImageView(Element elem, String sessionId,
			String baseUrl, CryptoUtils cryptoUtils)
	{
		super(elem);
		this.cryptoUtils = cryptoUtils;
		this.baseUrl = baseUrl;
		this.sessionId = sessionId;

		initialize(elem);

		attributes = getStyleSheet().getViewAttributes(this);
	}


	@Override
	public float getAlignment(int axis)
	{
		assert axis == X_AXIS || axis == Y_AXIS : "Illegal value of axis: "
				+ axis;

		if (axis == View.Y_AXIS) {
			return 1.0f;
		}
		return super.getAlignment(axis);
	}


	@Override
	public AttributeSet getAttributes()
	{
		return attributes;
	}


	@Override
	public void setParent(View parent)
	{
		super.setParent(parent);
		fContainer = (parent != null) ? getContainer() : null;

		if ((parent == null) && (fComponent != null)) {
			fComponent.getParent().remove(fComponent);
			fComponent = null;
		}
	}


	@Override
	public float getPreferredSpan(int axis)
	{
		int extra = 2 * (getBorder() + getSpace(axis));

		switch (axis) {
			case View.X_AXIS:
				return fWidth + extra;

			case View.Y_AXIS:
				return fHeight + extra;

			default:
				throw new IllegalArgumentException("Invalid axis: " + axis);
		}
	}


	@Override
	public void setSize(float width, float height)
	{
		//see javadoc comment
	}


	@Override
	public void changedUpdate(DocumentEvent e, Shape a, ViewFactory f)
	{
		super.changedUpdate(e, a, f);

		int height = fHeight;
		int width = fWidth;

		initialize(getElement());

		boolean hChanged = fHeight != height;
		boolean wChanged = fWidth != width;

		if (hChanged || wChanged) {
			preferenceChanged(null, hChanged, wChanged);
		}
	}


	// --- Progressive display ---------------------------------------------
	// This can come on any thread. If we are in the process of reloading
	// the image and determining our state (loading == true) we don't fire
	// preference changed, or repaint, we just reset the fWidth/fHeight as
	// necessary and return. This is ok as we know when loading finishes
	// it will pick up the new height/width, if necessary.
	public boolean imageUpdate(Image img, int flags, int x, int y, int width,
			int height)
	{
		if ((fImage == null) || (fImage != img)) {
			return false;
		}

		if ((flags & (ABORT | ERROR)) != 0) {
			fImage = null;
			repaint(0);

			return false;
		}

		// Resize image if necessary:
		short changed = 0;

		if ((flags & ImageObserver.HEIGHT) != 0) {
			if (!getElement().getAttributes().isDefined(HTML.Attribute.HEIGHT)) {
				changed |= 1;
			}
		}

		if ((flags & ImageObserver.WIDTH) != 0) {
			if (!getElement().getAttributes().isDefined(HTML.Attribute.WIDTH)) {
				changed |= 2;
			}
		}

		synchronized (this) {
			if ((changed & 1) == 1) {
				fWidth = width;
			}

			if ((changed & 2) == 2) {
				fHeight = height;
			}

			if (loading) {
				// No need to resize or repaint, still in the process of
				// loading.
				return true;
			}
		}

		if (changed != 0) {
			// May need to resize myself, asynchronously:
			Document doc = getDocument();

			try {
				if (doc instanceof AbstractDocument) {
					((AbstractDocument) doc).readLock();
				}

				preferenceChanged(null, true, true);
			}
			finally {
				if (doc instanceof AbstractDocument) {
					((AbstractDocument) doc).readUnlock();
				}
			}

			return true;
		}

		// Repaint when done or when new pixels arrive:
		if ((flags & (FRAMEBITS | ALLBITS)) != 0) {
			repaint(0);
		}
		else if ((flags & SOMEBITS) != 0) {
			if (sIsInc) {
				repaint(sIncRate);
			}
		}

		return ((flags & ALLBITS) == 0);
	}


	@Override
	public Shape modelToView(int pos, Shape a, Position.Bias b)
	{
		int p0 = getStartOffset();
		int p1 = getEndOffset();

		if ((pos >= p0) && (pos <= p1)) {
			Rectangle r = a.getBounds();

			if (pos == p1) {
				r.x += r.width;
			}

			r.width = 0;

			return r;
		}

		return null;
	}


	/**
	 * On double-click, open image properties dialog.
	 * @param e
	 *
	 * @todo Method not implemented.
	 */
	public void mouseClicked(MouseEvent e)
	{
		//        if (e.getClickCount() == 2) {
		//            //$ IMPLEMENT
		//        }
	}


	/**
	 * Resize image if initial click was in grow-box:
	 * @param e
	 */
	public void mouseDragged(MouseEvent e)
	{
		if (fGrowBase != null) {
			Point loc = fComponent.getLocationOnScreen();
			int width = Math.max(2, (loc.x + e.getX()) - fGrowBase.x);
			int height = Math.max(2, (loc.y + e.getY()) - fGrowBase.y);

			if (e.isShiftDown() && (fImage != null)) {
				// Make sure size is proportional to actual image size:
				float imgWidth = fImage.getWidth(this);
				float imgHeight = fImage.getHeight(this);

				if ((imgWidth > 0) && (imgHeight > 0)) {
					float prop = imgHeight / imgWidth;
					float pwidth = height / prop;
					float pheight = width * prop;

					if (pwidth > width) {
						width = (int) pwidth;
					}
					else {
						height = (int) pheight;
					}
				}
			}

			resize(width, height);
		}
	}


	public void mouseEntered(MouseEvent e)
	{
		//nothing to do for now.
	}


	public void mouseExited(MouseEvent e)
	{
		//nothing to do for now.
	}


	public void mouseMoved(MouseEvent e)
	{
		//nothing to do for now.
	}


	public void mousePressed(MouseEvent e)
	{
		Dimension size = fComponent.getSize();

		if ((e.getX() >= (size.width - 7)) && (e.getY() >= (size.height - 7))
				&& (getSelectionState() == 2)) {
			// Click in selected grow-box:
			Point loc = fComponent.getLocationOnScreen();
			fGrowBase = new Point((loc.x + e.getX()) - fWidth, (loc.y + e
					.getY())
					- fHeight);
		}
		else {
			// Else select image:
			fGrowBase = null;

			JTextComponent comp = (JTextComponent) fContainer;
			int start = fElement.getStartOffset();
			int end = fElement.getEndOffset();
			int mark = comp.getCaret().getMark();
			int dot = comp.getCaret().getDot();

			if (e.isShiftDown()) {
				// extend selection if shift key down:
				if (mark <= start) {
					comp.moveCaretPosition(end);
				}
				else {
					comp.moveCaretPosition(start);
				}
			}
			else {
				// just select image, without shift:
				if (mark != start) {
					comp.setCaretPosition(start);
				}

				if (dot != end) {
					comp.moveCaretPosition(end);
				}
			}
		}
	}


	/**
	 * @param e
	 * @todo Method not implemented.
	 */
	public void mouseReleased(MouseEvent e)
	{
		fGrowBase = null;

		//! Should post some command to make the action undo-able
	}


	// --- Painting --------------------------------------------------------

	@Override
	public void paint(Graphics g, Shape a)
	{
		Color oldColor = g.getColor();
		Rectangle alloc = (a instanceof Rectangle) ? (Rectangle) a : a
				.getBounds();
		fBounds.setBounds(alloc);

		int border = getBorder();
		int x = fBounds.x + border + getSpace(X_AXIS);
		int y = fBounds.y + border + getSpace(Y_AXIS);
		int width = fWidth;
		int height = fHeight;
		int sel = getSelectionState();

		// Make sure my Component is in the right place:

		/*
		 * if( fComponent == null ) { fComponent = new Component() { };
		 * fComponent.addMouseListener(this);
		 * fComponent.addMouseMotionListener(this);
		 * fComponent.setCursor(Cursor.getDefaultCursor()); // use arrow cursor
		 * fContainer.add(fComponent); } fComponent.setBounds(x,y,width,height);
		 */

		// If no pixels yet, draw gray outline and icon:
		if (!hasPixels(this)) {
			g.setColor(Color.lightGray);
			g.drawRect(x, y, width - 1, height - 1);
			g.setColor(oldColor);
			loadIcons();

			Icon icon = (fImage == null) ? sMissingImageIcon
					: sPendingImageIcon;

			if (icon != null) {
				icon.paintIcon(getContainer(), g, x, y);
			}
		}

		// Draw image:
		if (fImage != null) {
			g.drawImage(fImage, x, y, width, height, this);

			// Use the following instead of g.drawImage when
			// BufferedImageGraphics2D.setXORMode is fixed (4158822).
			//  Use Xor mode when selected/highlighted.
			//! Could darken image instead, but it would be more expensive.

			/*
			 * if( sel > 0 ) g.setXORMode(Color.white); g.drawImage(fImage,x, y,
			 * width,height,this); if( sel > 0 ) g.setPaintMode();
			 */
		}

		// If selected exactly, we need a black border & grow-box:
		Color bc = getBorderColor();

		if (sel == 2) {
			// Make sure there's room for a border:
			int delta = 2 - border;

			if (delta > 0) {
				x += delta;
				y += delta;
				width -= (delta << 1);
				height -= (delta << 1);
				border = 2;
			}

			bc = null;
			g.setColor(Color.black);

			// Draw grow box:
			g.fillRect((x + width) - 5, (y + height) - 5, 5, 5);
		}

		// Draw border:
		if (border > 0) {
			if (bc != null) {
				g.setColor(bc);
			}

			// Draw a thick rectangle:
			for (int i = 1; i <= border; i++) {
				g.drawRect(x - i, y - i, width - 1 + i + i, height - 1 + i + i);
			}

			g.setColor(oldColor);
		}
	}


	@Override
	public int viewToModel(float x, float y, Shape a, Position.Bias[] bias)
	{
		Rectangle alloc = (Rectangle) a;

		if (x < (alloc.x + alloc.width)) {
			bias[0] = Position.Bias.Forward;

			return getStartOffset();
		}

		bias[0] = Position.Bias.Backward;

		return getEndOffset();
	}


	private boolean isEditable()
	{
		return fContainer instanceof JEditorPane
				&& ((JEditorPane) fContainer).isEditable();
	}


	/**
	 * Determines whether the image is selected, and if it's the only thing
	 * selected.
	 *
	 * @return 0 if not selected, 1 if selected, 2 if exclusively selected.
	 *         "Exclusive" selection is only returned when editable.
	 */
	protected int getSelectionState()
	{
		int p0 = fElement.getStartOffset();
		int p1 = fElement.getEndOffset();

		if (fContainer instanceof JTextComponent) {
			JTextComponent textComp = (JTextComponent) fContainer;
			int start = textComp.getSelectionStart();
			int end = textComp.getSelectionEnd();

			if ((start <= p0) && (end >= p1)) {
				if ((start == p0) && (end == p1) && isEditable()) {
					return 2;
				}
				return 1;
			}
		}

		return 0;
	}


	private StyleSheet getStyleSheet()
	{
		HTMLDocument doc = (HTMLDocument) getDocument();

		return doc.getStyleSheet();
	}


	private void repaint(long delay)
	{
		if ((fContainer != null) && (fBounds != null)) {
			fContainer.repaint(delay, fBounds.x, fBounds.y, fBounds.width,
				fBounds.height);
		}
	}


	private void resize(int width, int height)
	{
		if ((width == fWidth) && (height == fHeight)) {
			return;
		}

		fWidth = width;
		fHeight = height;

		// Replace attributes in document:
		MutableAttributeSet attr = new SimpleAttributeSet();
		attr.addAttribute(HTML.Attribute.WIDTH, Integer.toString(width));
		attr.addAttribute(HTML.Attribute.HEIGHT, Integer.toString(height));
		((StyledDocument) getDocument()).setCharacterAttributes(fElement
				.getStartOffset(), fElement.getEndOffset(), attr, false);
	}


	private int getBorder()
	{
		return border;
	}


	private Color getBorderColor()
	{
		StyledDocument doc = (StyledDocument) getDocument();

		return doc.getForeground(getAttributes());
	}


	private boolean isLink()
	{
		return isLink;
	}


	private int getSpace(int axis)
	{
		if (axis == X_AXIS) {
			return xSpace;
		}
		return ySpace;
	}


	private boolean hasPixels(ImageObserver obs)
	{
		return (fImage != null) && (fImage.getHeight(obs) > 0)
				&& (fImage.getWidth(obs) > 0);
	}


	private int getIntAttr(HTML.Attribute name, int deflt)
	{
		AttributeSet attr = fElement.getAttributes();

		if (attr.isDefined(name)) { // does not check parents!

			int i;
			String val = (String) attr.getAttribute(name);

			if (val == null) {
				i = deflt;
			}
			else {
				try {
					i = Math.max(0, Integer.parseInt(val));
				}
				catch (NumberFormatException x) {
					i = deflt;
				}
			}

			return i;
		}
		return deflt;
	}


	private URL getSourceURL()
	{
		String src = (String) fElement.getAttributes().getAttribute(
			HTML.Attribute.SRC);

		if (src == null) {
			return null;
		}

		try {
			URL reference = new URL(baseUrl);
			// Add a sessionid to the request if necessary

			if (src.indexOf(";") < 0) {
				if (src.indexOf('?') < 0) {
					src = src + ";jsessionid=" + sessionId;
				}
				else {
					src = src.substring(0, src.indexOf('?')) + ";jsessionid="
							+ sessionId + src.substring(src.indexOf('?'));
				}
			}

			URL u = new URL(reference, src);
			return u;
		}
		catch (MalformedURLException e) {
			return null;
		}
	}


	private void initialize(Element elem)
	{
		synchronized (this) {
			loading = true;
			fWidth = 0;
			fHeight = 0;
		}

		int width = 0;
		int height = 0;
		boolean customWidth = false;
		boolean customHeight = false;

		try {
			fElement = elem;

			AttributeSet attr = elem.getAttributes();

			AttributeSet anchorAttr = (AttributeSet) attr
					.getAttribute(HTML.Tag.A);
			isLink = ((anchorAttr != null) && anchorAttr
					.isDefined(HTML.Attribute.HREF));
			border = getIntAttr(HTML.Attribute.BORDER,
				isLink() ? DEFAULT_BORDER : 0);
			xSpace = getIntAttr(HTML.Attribute.HSPACE, 0);
			ySpace = getIntAttr(HTML.Attribute.VSPACE, 0);

			// Request image from document's cache:
			URL src = getSourceURL();

			if (src != null) {
				Dictionary cache = (Dictionary) getDocument().getProperty(
					IMAGE_CACHE_PROPERTY);

				if (cache != null) {
					fImage = (Image) cache.get(src);
				}
				else {
					if (cryptoUtils != null) {
						// Decipher the image
						try {
							// Get an input stream from the url
							HttpURLConnection conn = (HttpURLConnection) src
									.openConnection();
							conn.setAllowUserInteraction(false);
							conn.setDoOutput(false);

							InputStream in = cryptoUtils
									.getDecipheredInputStream(conn.getInputStream());

							// Store the deciphered data in an array
							ByteArrayOutputStream out = new ByteArrayOutputStream();
							byte[] buff = new byte[4096];
							int bytesRead;

							while ((bytesRead = in.read(buff)) > 0) {
								out.write(buff, 0, bytesRead);
							}

							in.close();
							out.close();

							// Create the image
							fImage = Toolkit.getDefaultToolkit().createImage(
								out.toByteArray());
						}
						catch (Exception e) {
							e.printStackTrace();
						}
					}
					else {
						fImage = Toolkit.getDefaultToolkit().getImage(src);
					}
				}
			}

			// Get height/width from params or image or defaults:
			height = getIntAttr(HTML.Attribute.HEIGHT, -1);
			customHeight = (height > 0);

			if (!customHeight && (fImage != null)) {
				height = fImage.getHeight(this);
			}

			if (height <= 0) {
				height = DEFAULT_HEIGHT;
			}

			width = getIntAttr(HTML.Attribute.WIDTH, -1);
			customWidth = (width > 0);

			if (!customWidth && (fImage != null)) {
				width = fImage.getWidth(this);
			}

			if (width <= 0) {
				width = DEFAULT_WIDTH;
			}

			// Make sure the image starts loading:
			if (fImage != null) {
				if (customWidth && customHeight) {
					Toolkit.getDefaultToolkit().prepareImage(fImage, height,
						width, this);
				}
				else {
					Toolkit.getDefaultToolkit().prepareImage(fImage, -1, -1,
						this);
				}
			}
		}
		finally {
			synchronized (this) {
				loading = false;

				if (customWidth || (fWidth == 0)) {
					fWidth = width;
				}

				if (customHeight || (fHeight == 0)) {
					fHeight = height;
				}
			}
		}
	}


	private void loadIcons()
	{
		try {
			synchronized (this.getClass()) {
				if (sPendingImageIcon == null) {
					sPendingImageIcon = makeIcon(PENDING_IMAGE_SRC);
				}

				if (sMissingImageIcon == null) {
					sMissingImageIcon = makeIcon(MISSING_IMAGE_SRC);
				}
			}
		}
		catch (Exception x) {
			x.printStackTrace();
		}
	}


	// --- Static icon accessors -------------------------------------------
	private Icon makeIcon(final String gifFile)
		throws IOException
	{
		/*
		 * Copy resource into a byte array. This is necessary because several
		 * browsers consider Class.getResource a security risk because it can be
		 * used to load additional classes. Class.getResourceAsStream just
		 * returns raw bytes, which we can convert to an image.
		 */

		//	InputStream resource = HTMLEditorKit.getResourceAsStream(gifFile);
		InputStream resource = getClass().getResourceAsStream(gifFile);

		if (resource == null) {
			return null;
		}
		ByteArrayOutputStream out = new ByteArrayOutputStream(1024);
		byte[] buffer = new byte[1024];
		int n;

		BufferedInputStream in = null;

		try {
			in = new BufferedInputStream(resource);
			while ((n = in.read(buffer)) > 0) {
				out.write(buffer, 0, n);
			}
		}
		finally {
			if (in != null) {
				in.close();
				out.flush();
			}
		}

		buffer = out.toByteArray();

		if (buffer.length == 0) {
			return null;
		}

		return new ImageIcon(buffer);
	}
}