/**
 * $Id: OperationInProgressDialog.java,v 1.2 2007-09-25 15:15:27 maneo Exp $
 *
 * Copyright (c) 2002 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.ps.applet.operation;

import java.awt.Dialog;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.WindowConstants;

/**
 * Dialog used to inform user that there is operation in progress.
 */
public class OperationInProgressDialog 
	extends JDialog
{
	private static final String PROGRESSLABEL = 
		"Prosz czeka\u0107...";
	private static final String TITLE =
		"Trwa wykonywanie operacji";
	
	private JPanel innerPanel = null; 
	private JLabel progressLabel = null;
	private JProgressBar progressBar = null;
	
	/**
	 * Returns new instance of OperationInProgressDialog. 
	 */
	public OperationInProgressDialog() 
	{
		init();
	}
	
    /**
     * Returns new instance of OperationInProgressDialog. 
     * @param parent parent frame
     */    
	public OperationInProgressDialog(Frame parent)
    {
        super(parent);
        init();        
    }

    /**
     * Returns new instance of OperationInProgressDialog. 
     * @param parent parent dialog
     */
    public OperationInProgressDialog(Dialog parent)
    {
        super(parent);
        init();
    }

    private void init()
    {
        setSize(250, 100);
        setTitle(TITLE);
        setLocation(
                (Toolkit.getDefaultToolkit()
                        .getScreenSize().width - getWidth()) / 2,
                (Toolkit.getDefaultToolkit()
                        .getScreenSize().height - getHeight()) / 2);
        setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        getContentPane().setLayout(new GridBagLayout());
        GridBagConstraints cons = new GridBagConstraints();
        cons.weightx = 1;
        cons.weighty = 1;
        cons.fill = GridBagConstraints.BOTH;
        getContentPane().add(getInnerPanel(), cons);
        initGUITexts();
    }
	
	/**
	 * Sets new window parent for the dialog.
	 * @param parent Parent window.
	 */
	public void setParent(Window parent) 
	{
		parent.addWindowListener(
			new WindowAdapter() {
				@Override
                public void windowActivated(WindowEvent e) {					
					toFront();
				}
			});
	}
	
	private void initGUITexts()
	{
		setTitle("Operacja w toku");
		getProgressLabel().setText("Prosz\u0119 czeka\u0107...");
	}
	
	private JPanel getInnerPanel()
	{
		if (innerPanel == null) {
			innerPanel = new JPanel();
			innerPanel.setLayout(new GridBagLayout());
			
			GridBagConstraints cons = new GridBagConstraints();
			cons.fill = GridBagConstraints.HORIZONTAL;
			cons.weightx = 1;
			cons.gridy = 0;
			cons.insets = new Insets(0, 8, 0, 8);
			innerPanel.add(getProgressBar(), cons);
			
			cons = new GridBagConstraints();
			cons.gridy = 1;
			innerPanel.add(getProgressLabel(), cons);
		}
		return innerPanel;
	}
	
	private JProgressBar getProgressBar()
	{
		if (progressBar == null) {
			progressBar = new JProgressBar();
			progressBar.setIndeterminate(true);
		}
		return progressBar;
	}
	
	private JLabel getProgressLabel()
	{
		if (progressLabel == null) {
			progressLabel = new JLabel();			
			progressLabel.setText(PROGRESSLABEL);
		}
		return progressLabel;
	}
}
