/**
 * $Id: OperationInProgressManager.java,v 1.1 2006-05-16 07:59:38 tparkola Exp $
 *
 * Copyright (c) 2002 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.ps.applet.operation;

import java.awt.Dialog;
import java.awt.Frame;
import java.awt.Window;

import javax.swing.SwingUtilities;

/**
 * Manages operation in progress functionality.
 */
public class OperationInProgressManager 
{
	private OperationInProgressDialog dialog;
	
	/**
	 * Returns new instance of OperationInProgressManager. 
	 * @param parent Parent window for the operation in progress dialog.
	 */
	public OperationInProgressManager(final Window parent)
	{
        if (parent instanceof Frame) {
            dialog = new OperationInProgressDialog((Frame)parent);
        }
        else if (parent instanceof Dialog) {
            dialog = new OperationInProgressDialog((Dialog)parent);
        }
        else {
            dialog = new OperationInProgressDialog();
        }	    
        dialog.setParent(parent);
		dialog.setModal(true);	
	}
    
    /**
     * Returns new instance of OperationInProgressManager. 
     * @param parent Parent frame for the operation in progress dialog.
     */
    public OperationInProgressManager(final Frame parent)
    {
        dialog = new OperationInProgressDialog(parent);
        dialog.setModal(true);  
    } 
    
    /**
     * Returns new instance of OperationInProgressManager. 
     * @param parent Parent dialog for the operation in progress dialog.
     */
    public OperationInProgressManager(final Dialog parent)
    {
        dialog = new OperationInProgressDialog(parent);
        dialog.setModal(true);  
    }     
	
	/**
	 * Starts the operation in progress dialog working.
	 */
	public void start()
	{
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run()
            {
                dialog.setVisible(true);                
            }            
        });
	}
	
	/**
	 * Finishes the operation in progress dialog working.
	 */
	public void stop()
	{
        SwingUtilities.invokeLater(new Runnable() {
            public void run()
            {
                dialog.dispose();
            }            
        });                          						
	}

    /**
     * Creates thread and runs a given task in it.
     * @param progress the task to perform
     * @param frame the parent of the operation in progress dialog
     */
    public static void runTask(final OperationInProgress progress, 
            final Window frame)
    {
        final OperationInProgressManager manager = 
            new OperationInProgressManager(frame);
        new Thread(new Runnable() {        
            public void run()
            {
                try {
                    progress.doOperation(manager);
                }
                catch (Exception e) {
                    manager.stop();
                }
            }        
        }).start();
    }
    
    /**
     * Creates thread and runs a given task in it.
     * @param progress the task to perform
     * @param dialog the parent of the operation in progress dialog
     */
    public static void runTask(final OperationInProgress progress, 
            final Dialog dialog)
    {
        final OperationInProgressManager manager = 
            new OperationInProgressManager(dialog);
        new Thread(new Runnable() {        
            public void run()
            {
                progress.doOperation(manager);                
            }        
        }).start();
    }    
}
