/*
 * Copyright (c) 2000 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */

package pl.psnc.dlibra.ps.applet.viewer;

import java.util.*;


/**
 * Implements a browser-like history buffer
 * @version $Id: History.java,v 1.3 2006-05-29 13:36:09 mwerla Exp $
 */
public class History
{
    /** Stores back entries */
    private Stack<String> backStack;

    /** Stores forward entries */
    private Stack<String> forwardStack;

    /** Stores current entry */
    private String currentEntry;

    /**
     * Creates a new History.
     */
    public History()
    {
        backStack = new Stack<String>();
        forwardStack = new Stack<String>();
        currentEntry = null;
    }

    /**
     * Indicates if the "back" movement is possible.
     *
     * @return boolean true when the "back" movement is possible.
     */
    public boolean isBackEnabled()
    {
        return !backStack.isEmpty();
    }


    /**
     * Indicates if the "forward" movement is possible.
     *
     * @return boolean true when the "forward" movement is possible.
     */
    public boolean isForwardEnabled()
    {
        return !forwardStack.isEmpty();
    }


    /**
     * Moves "back" in the history.
     *
     * @return String the entry that was "back" in the history (e.g. the
     *         previous page)
     */
    public String back()
    {
        if (!backStack.isEmpty()) {
            String entry = backStack.pop().toString();
            forwardStack.push(currentEntry);
            currentEntry = entry;

            return entry.toString();
        }
        return null;
    }


    /**
     * Moves "forward" in the history.
     *
     * @return String the entry that was "next" in the history (e.g. the next
     *         page)
     */
    public String forward()
    {
        if (!forwardStack.isEmpty()) {
            String entry = forwardStack.pop().toString();
            backStack.push(currentEntry);
            currentEntry = entry;

            return entry.toString();
        }
        return null;
    }


    /**
     * Puts a new entry to the history buffer and discards all "forward"
     * history.
     *
     * @param entry the new entry to be put in the history buffer.
     */
    public void put(String entry)
    {
        if (currentEntry != null) {
            backStack.push(currentEntry);
        }

        currentEntry = entry;

        forwardStack.clear();
    }
}
