/*
 * Copyright (c) 2000 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */

/*
 * PublicationViewer.java
 *
 * Created on 15 April 2002, 11:45
 */
package pl.psnc.dlibra.ps.applet.viewer;

import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.security.Security;

import javax.crypto.SecretKey;
import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import javax.swing.JApplet;
import javax.swing.SwingUtilities;

import pl.psnc.dlibra.ps.crypto.KeyAgreementRequestPayload;
import pl.psnc.dlibra.ps.crypto.KeyAgreementResponsePayload;
import pl.psnc.dlibra.ps.crypto.KeyAgreementServer;

/**
 * An applet for viewing secured dLibra publications.
 * @version $Id: PublicationViewer.java,v 1.9 2007-10-30 11:15:31 tparkola Exp $
 */
public class PublicationViewer
	extends JApplet
{

	private static final long serialVersionUID = -4871305195063164918L;

	private static final String HTTPS = "https://";

	private static final String HTTP = "http://";

	/** Panel containing GUI */
	private pl.psnc.dlibra.ps.applet.viewer.PublicationViewerPanel publicationViewerPanel;

	/** Session secret key */
	private SecretKey key;

	/** A path to the pl.psnc.dlibra servlets. Ends with '/' */
	private String serverPath;

	/** Current HTTP sessionId */
	private String sessionId;


	@Override
	public void init()
	{
		super.init();

		this.sessionId = getParameter("sessionId");
		this.serverPath = getParameter("initialUrl").substring(0,
			getParameter("initialUrl").lastIndexOf("/Content/"))
				+ "/";

		publicationViewerPanel = new pl.psnc.dlibra.ps.applet.viewer.PublicationViewerPanel();
		publicationViewerPanel.setLang(getParameter("userLang"));
		publicationViewerPanel.setInitialParameters(getParameter("initialUrl"),
			getParameter("sessionId"));

		getContentPane().add(publicationViewerPanel,
			java.awt.BorderLayout.CENTER);

		new Thread(new Runnable() {

			public void run()
			{
				try {
					key = performKeyAgreement(serverPath + "ka;jsessionid="
							+ sessionId);
					publicationViewerPanel.setKey(key);

					javax.swing.SwingUtilities.invokeLater(new Runnable() {

						public void run()
						{
							publicationViewerPanel.start();
						}
					});
				}
				catch (final Exception e) {
					// Display the message in the browser window
					SwingUtilities.invokeLater(new Runnable() {

						public void run()
						{
							publicationViewerPanel.showErrorPage(e);
						}
					});
				}
			}
		}).start();
	}


	private SecretKey performKeyAgreement(String targetUrl)
		throws Exception
	{
		// Send the first request
		URL url = new URL(targetUrl);
		HttpURLConnection conn;

		conn = (HttpURLConnection) url.openConnection();
		conn.setRequestMethod("POST");
		conn.setAllowUserInteraction(false);
		conn.setDoOutput(true);
		conn.setRequestProperty("Content-type", "binary/x-java-serialized");
		conn.getOutputStream().close();

		// Get the response
		conn.connect();

		InputStream is = conn.getInputStream();
		ObjectInputStream ois = null;
		ois = new ObjectInputStream(is);

		KeyAgreementRequestPayload karp = (KeyAgreementRequestPayload) ois
				.readObject();
		ois.close();

		// Do the first phase
		KeyAgreementServer kas = new KeyAgreementServer();
		KeyAgreementResponsePayload karep;

		karep = kas.process(karp);

		// Send the results to the client
		conn = (HttpURLConnection) url.openConnection();
		conn.setRequestMethod("POST");
		conn.setAllowUserInteraction(false);
		conn.setDoOutput(true);
		conn.setRequestProperty("Content-type", "binary/x-java-serialized");

		ObjectOutputStream oos = new ObjectOutputStream(conn.getOutputStream());
		oos.writeObject(karep);
		oos.flush();
		oos.close();

		// We don't read any input here
		conn.connect();
		conn.getInputStream().close();

		// Do the second phase
		return kas.finish();
	}

}
