/**
 * $Id: PublicationViewerContentManager.java,v 1.9 2006-07-31 12:58:55 mwerla Exp $
 *
 * Copyright (c) 2002 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.ps.applet.viewer;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.URLConnection;

import pl.psnc.dlibra.ps.applet.VersionProvider;
import pl.psnc.util.crypto.CryptoUtils;

/**
 *  @version $Id: PublicationViewerContentManager.java,v 1.9 2006-07-31 12:58:55 mwerla Exp $
 */
public class PublicationViewerContentManager
{

	private static final String CONTENT_HTML = "text/html";

	private String[][] imageContentTypes = { { "image/jpg", "jpg"},
			{ "image/jpeg", "jpeg"}, { "image/gif", "gif"},
			{ "image/png", "png"}, { "image/bmp", "bmp"}};

	private static final String[] metaTags = { "<META ", "<meta ", "<Meta "};

	private String currentUrl = null;

	private History history = new History();

	private CryptoUtils cryptoUtils;

	private boolean embeddedImage = false;


	/**
	 * @param utils Crypto utils to decipher content
	 */
	public PublicationViewerContentManager(CryptoUtils utils)
	{
		cryptoUtils = utils;
	}


	/**
	 * @return the current URL
	 */
	public String getCurrentUrl()
	{
		return currentUrl;
	}


	/**
	 * @return History
	 */
	public History getHistory()
	{
		return history;
	}


	/**
	 * @param url Content URL
	 * @return Content bean with requested content.
	 * @throws IOException
	 */
	public ContentBean getDocument(String url)
		throws IOException
	{
		String contentType = null;
		ContentBean result = new ContentBean();
		String extension = getExtension(url);
		for (int i = 0; i < imageContentTypes.length; i++) {
			if (extension.toLowerCase().startsWith(imageContentTypes[i][1])) {
				contentType = imageContentTypes[i][0];
			}
		}

		// Check if the link points to an image
		boolean image = false;

		if (contentType != null) {
			for (int i = 0; i < imageContentTypes.length; i++) {
				if (contentType.equalsIgnoreCase(imageContentTypes[i][0])) {
					image = true;
					break;
				}
			}
		}

		result.setContentType(CONTENT_HTML);
		currentUrl = url;

		// Wrap the image with some HTML code
		if (image) {
			embeddedImage = true;
			result.setContent("<html><head><title>" + url
					+ "</title></head><body><img src='" + url
					+ "'></body></html>");
		}
		else {
			if (embeddedImage) {
				result.setContentType(CONTENT_HTML);
			}
			embeddedImage = false;
			result.setContent(getDocument(new URL(url)));
		}
		return result;
	}


	private String getExtension(String url)
	{
		String extension = url;

		if (extension.lastIndexOf(";") > -1) {
			extension = extension.substring(0, extension.lastIndexOf(";"));
		}
		extension = extension.substring(extension.lastIndexOf("/") + 1);

		if (extension.lastIndexOf("#") > -1) {
			extension = extension.substring(0, extension.lastIndexOf("#"));
		}
		extension = extension.substring(extension.lastIndexOf(".") + 1);
		return extension;
	}


	private String getDocument(URL doc)
		throws IOException
	{
		URLConnection connection = doc.openConnection();
		connection.setRequestProperty("User-Agent", "dLibra-html-plugin/"
				+ VersionProvider.getVersion() + " (http://dlibra.psnc.pl/)");

		InputStream encInputStream = cryptoUtils
				.getDecipheredInputStream(connection.getInputStream());
		String docContent = "";

		String encoding = "UTF-8";

		BufferedReader reader = new BufferedReader(new InputStreamReader(
				encInputStream, encoding));

		String lineRead = null;
		try {
			while ((lineRead = reader.readLine()) != null) {
				if (!isMetaTagLine(lineRead)) {
					docContent = docContent + lineRead + "\n";
				}
			}
		}
		finally {
			encInputStream.close();
		}

		return docContent;
	}


	private boolean isMetaTagLine(String lineRead)
	{
		boolean metaFound = false;
		for (int i = 0; i < metaTags.length; i++) {
			if (lineRead.indexOf(metaTags[i]) >= 0) {
				metaFound = true;
				break;
			}
		}
		return metaFound;
	}

}