/*
 * Copyright (c) 2000 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */

/*
 * PublicationViewerPanel.java
 *
 * Created on 12 April 2002, 15:21
 */
package pl.psnc.dlibra.ps.applet.viewer;

import java.awt.EventQueue;
import java.awt.Point;
import java.awt.Rectangle;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InvalidClassException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.URL;
import java.util.InvalidPropertiesFormatException;
import java.util.Properties;

import javax.crypto.SecretKey;
import javax.swing.Action;
import javax.swing.ImageIcon;
import javax.swing.JCustomEditorPane;
import javax.swing.JEditorPane;
import javax.swing.JPanel;
import javax.swing.JViewport;
import javax.swing.SwingUtilities;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.text.ViewerStyledEditorKit;
import javax.swing.text.html.CustomImportStyleSheet;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.html.HTMLFrameHyperlinkEvent;
import javax.swing.text.html.ViewerHTMLEditorKit;

import pl.psnc.dlibra.ps.applet.operation.OperationInProgress;
import pl.psnc.dlibra.ps.applet.operation.OperationInProgressManager;
import pl.psnc.util.crypto.CryptoUtils;

/**
 * @version $Id: PublicationViewerPanel.java,v 1.14 2007-04-23 08:26:02 mwerla Exp $
 */
public class PublicationViewerPanel
	extends JPanel
	implements HyperlinkListener
{

	private static final long serialVersionUID = -1323296741781693729L;

	private static final String CONTENT_HTML = "text/html";

	private static final String PROTOCOL_HTTP = "http://";

	private static final String CLIPBOARD_CUT = "cut-to-clipboard";

	private static final String CLIPBOARD_COPY = "copy-to-clipboard";

	private static final String INVALID_CLASS_ERROR = "Not String nor URL class returned.";

	private PublicationViewerContentManager contentManager;

	private javax.swing.JButton backButton;

	private javax.swing.JButton forwardButton;

	private javax.swing.JButton goButton;

	private javax.swing.JCustomEditorPane mainEditorPane;

	private javax.swing.JLabel addressLabel;

	private javax.swing.JLabel statusLabel;

	private javax.swing.JPanel progressPanel;

	private javax.swing.JProgressBar progressBar;

	private javax.swing.JPanel navigationPanel;

	private javax.swing.JScrollPane mainScrollPane;

	private javax.swing.JTextField addressTextField;

	private String initialUrl;

	private static final String SESSION_ID = "jsessionid";

	private String sessionId;

	private CryptoUtils utils;

	private Properties lang;


	/**
	 * Creates new form PublicationViewerPanel
	 */
	public PublicationViewerPanel()
	{
		initComponents();
		setLang("pl");
		mainEditorPane.addHyperlinkListener(this);

		// Prepare editor pane
		ViewerHTMLEditorKit editorKit = new ViewerHTMLEditorKit();
		mainEditorPane.setEditorKitForContentType(CONTENT_HTML, editorKit);

		// Disable Cut/Copy
		Action[] actions = mainEditorPane.getActions();

		for (int i = 0; i < actions.length; i++) {
			if (actions[i].getValue(Action.NAME).equals(CLIPBOARD_CUT)
					|| actions[i].getValue(Action.NAME).equals(CLIPBOARD_COPY)) {

				actions[i].setEnabled(false);
			}
		}

		updateNavigation();

		addressTextField.setVisible(false);
		addressLabel.setVisible(false);

		statusLabel.setText(lang.getProperty("status_ready_java")
				+ System.getProperty("java.version") + ".");
	}


	/**
	 * Sets user language.
	 * @param userLang User language.
	 */
	void setLang(final String userLang)
	{
		loadLang(userLang);
		loadText();
	}


	private void loadLang(final String userLang)
	{
		lang = new Properties();
		try {
			lang.loadFromXML(getClass().getResourceAsStream(
				"/pl/psnc/dlibra/ps/applet/viewer/res/PublicationViewerPanel_"
						+ userLang + ".xml"));
		}
		catch (InvalidPropertiesFormatException e) {
			throw new RuntimeException(e);
		}
		catch (IOException e) {
			throw new RuntimeException(e);
		}
	}


	private void loadText()
	{
		goButton.setText(lang.getProperty("refresh"));
		backButton.setText(lang.getProperty("back"));
		forwardButton.setText(lang.getProperty("forward"));
		addressLabel.setText(lang.getProperty("address"));
		statusLabel.setText(lang.getProperty("status_ready"));
	}


	/**
	 * @param url initial url
	 * @param sessionId used by applet
	 */
	public void setInitialParameters(String url, String sessionId)
	{
		this.sessionId = sessionId;
		ViewerHTMLEditorKit.setSessionId(sessionId);

		showConnectingPage();

		// Add session id to the initial url
		initialUrl = getInitialUrl(url);
	}


	/**
	 * Sets secret key
	 * @param key
	 */
	public void setKey(SecretKey key)
	{
		utils = new CryptoUtils(key);
		contentManager = new PublicationViewerContentManager(utils);
		mainEditorPane.setCryptoUtils(utils);
		ViewerHTMLEditorKit.setCryptoUtils(utils);
		ViewerStyledEditorKit.setCryptoUtils(utils);
	}


	private final String getStatusError(String mess)
	{
		return lang.getProperty("error") + mess;
	}


	private final String getPageError(String title, String message)
	{
		return lang.getProperty("page_error.top") + lang.getProperty("error")
				+ lang.getProperty("page_error.first_inner") + title
				+ lang.getProperty("page_error.second_inner") + message
				+ lang.getProperty("page_error.down");
	}


	public void hyperlinkUpdate(javax.swing.event.HyperlinkEvent e)
	{
		if (e.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
			JEditorPane pane = (JEditorPane) e.getSource();

			if (e instanceof HTMLFrameHyperlinkEvent) {
				HTMLFrameHyperlinkEvent evt = (HTMLFrameHyperlinkEvent) e;
				HTMLDocument doc = (HTMLDocument) pane.getDocument();
				doc.processHTMLFrameHyperlinkEvent(evt);
				contentManager.getHistory().put(evt.getURL().toString());
				updateNavigation();
			}
			else {
				try {
					URL url = e.getURL();
					String newURL;
					if (url != null) {
						newURL = url.toString();
					}
					else {
						String previousURL = addressTextField.getText();
						int endPosition = previousURL.lastIndexOf('.') > previousURL
								.lastIndexOf('/') ? previousURL
								.lastIndexOf('/') : previousURL.length();

						previousURL = previousURL.substring(0, endPosition);
						newURL = e.getDescription();
						if (!(previousURL.endsWith("/") || newURL
								.startsWith("/"))) {
							newURL = previousURL + "/" + newURL;
						}
						else {
							newURL = previousURL + newURL;
						}
					}
					loadDocument(newURL);
					contentManager.getHistory().put(newURL);
					updateNavigation();
				}
				catch (Throwable t) {
					showErrorPage(t);
				}
			}
		}
	}


	/**
	 *
	 */
	public void showConnectingPage()
	{
		mainEditorPane.setContentType(CONTENT_HTML);
		mainEditorPane.setText(lang.getProperty("page_connecting"));
	}


	/**
	 * Handles errors. Shows user friendly error pages.
	 * @param e error to handle
	 */
	public void showErrorPage(Throwable e)
	{
		String title = "";
		String message = "";

		if (e instanceof FileNotFoundException) {
			title = lang.getProperty("file_not_found.title");
			message = lang.getProperty("file_not_found.message_first")
					+ e.getMessage()
					+ lang.getProperty("file_not_found.message_last");
		}
		else {
			title = e.getMessage();
			StringWriter stringWriter = new StringWriter();
			e.printStackTrace(new PrintWriter(stringWriter));
			message = "<pre>" + stringWriter.toString() + "</pre>";
		}
		String status = getStatusError(title);
		statusLabel.setText(status);
		mainEditorPane.setContentType(CONTENT_HTML);
		mainEditorPane.setText(getPageError(title, message));
	}


	/**
	 *
	 */
	public void start()
	{
		loadDocument(initialUrl);
		contentManager.getHistory().put(initialUrl);
		updateNavigation();
	}


	private void addressTextFieldActionPerformed(@SuppressWarnings("unused")
	java.awt.event.ActionEvent evt)
	{
		loadDocument(addressTextField.getText());
		contentManager.getHistory().put(addressTextField.getText());
		updateNavigation();
	}


	private void backButtonActionPerformed(@SuppressWarnings("unused")
	java.awt.event.ActionEvent evt)
	{
		loadDocument(contentManager.getHistory().back());
		updateNavigation();
	}


	private void forwardButtonActionPerformed(@SuppressWarnings("unused")
	java.awt.event.ActionEvent evt)
	{
		loadDocument(contentManager.getHistory().forward());
		updateNavigation();
	}


	private void goButtonActionPerformed(@SuppressWarnings("unused")
	java.awt.event.ActionEvent evt)
	{
		if (contentManager.getCurrentUrl() == null) {
			return;
		}

		mainEditorPane.setContentType(CONTENT_HTML);
		//mainEditorPane.setText("");

		loadDocument(contentManager.getCurrentUrl());
	}


	/**
	 * This method is called from within the constructor to initialize the
	 * form. WARNING: Do NOT modify this code. The content of this method is
	 * always regenerated by the Form Editor.
	 */
	private void initComponents()
	{
		addressTextField = new javax.swing.JTextField();
		addressLabel = new javax.swing.JLabel();
		statusLabel = new javax.swing.JLabel();
		mainScrollPane = new javax.swing.JScrollPane();
		mainEditorPane = new JCustomEditorPane();
		navigationPanel = new javax.swing.JPanel();
		backButton = new javax.swing.JButton();
		forwardButton = new javax.swing.JButton();
		goButton = new javax.swing.JButton();
		progressPanel = new javax.swing.JPanel();
		progressBar = new javax.swing.JProgressBar(0, 100);

		setLayout(new java.awt.GridBagLayout());

		java.awt.GridBagConstraints gridBagConstraints1;

		addressTextField.setFont(new java.awt.Font("Default", 0, 11));
		addressTextField.setText(PROTOCOL_HTTP);
		addressTextField.addActionListener(new java.awt.event.ActionListener() {

			public void actionPerformed(java.awt.event.ActionEvent evt)
			{
				addressTextFieldActionPerformed(evt);
			}
		});

		gridBagConstraints1 = new java.awt.GridBagConstraints();
		gridBagConstraints1.gridx = 1;
		gridBagConstraints1.gridy = 1;
		gridBagConstraints1.gridwidth = java.awt.GridBagConstraints.REMAINDER;
		gridBagConstraints1.fill = java.awt.GridBagConstraints.HORIZONTAL;
		gridBagConstraints1.insets = new java.awt.Insets(0, 0, 0, 5);
		gridBagConstraints1.weightx = 1.0;
		add(addressTextField, gridBagConstraints1);

		addressLabel.setFont(new java.awt.Font("Default", 0, 11));

		gridBagConstraints1 = new java.awt.GridBagConstraints();
		gridBagConstraints1.gridx = 0;
		gridBagConstraints1.gridy = 1;
		gridBagConstraints1.gridwidth = 1;
		gridBagConstraints1.gridheight = 1;
		gridBagConstraints1.fill = java.awt.GridBagConstraints.HORIZONTAL;
		gridBagConstraints1.insets = new java.awt.Insets(0, 5, 0, 3);
		gridBagConstraints1.weightx = 0.0;
		gridBagConstraints1.anchor = java.awt.GridBagConstraints.EAST;
		add(addressLabel, gridBagConstraints1);

		statusLabel.setFont(new java.awt.Font("Default", 0, 11));
		gridBagConstraints1 = new java.awt.GridBagConstraints();
		gridBagConstraints1.gridx = 0;
		gridBagConstraints1.gridy = 3;
		gridBagConstraints1.gridwidth = java.awt.GridBagConstraints.REMAINDER;
		gridBagConstraints1.fill = java.awt.GridBagConstraints.HORIZONTAL;
		gridBagConstraints1.insets = new java.awt.Insets(0, 5, 2, 5);
		gridBagConstraints1.weightx = 1.0;
		add(statusLabel, gridBagConstraints1);

		mainEditorPane.setEditable(false);
		mainScrollPane.setViewportView(mainEditorPane);

		gridBagConstraints1 = new java.awt.GridBagConstraints();
		gridBagConstraints1.gridx = 0;
		gridBagConstraints1.gridy = 2;
		gridBagConstraints1.gridwidth = java.awt.GridBagConstraints.REMAINDER;
		gridBagConstraints1.fill = java.awt.GridBagConstraints.BOTH;
		gridBagConstraints1.insets = new java.awt.Insets(3, 5, 3, 5);
		gridBagConstraints1.weightx = 1.0;
		gridBagConstraints1.weighty = 1.0;
		add(mainScrollPane, gridBagConstraints1);

		navigationPanel.setLayout(new java.awt.FlowLayout(
				java.awt.FlowLayout.LEFT, 5, 3));

		backButton.setIcon(getIconBack());
		backButton.setFont(new java.awt.Font("Default", 0, 11));
		backButton.setPreferredSize(new java.awt.Dimension(100, 19));
		backButton.setMaximumSize(new java.awt.Dimension(100, 19));
		backButton.setMargin(new java.awt.Insets(2, 2, 2, 2));
		backButton.setMinimumSize(new java.awt.Dimension(100, 19));
		backButton.addActionListener(new java.awt.event.ActionListener() {

			public void actionPerformed(java.awt.event.ActionEvent evt)
			{
				backButtonActionPerformed(evt);
			}
		});

		navigationPanel.add(backButton);

		forwardButton.setIcon(getIconForward());
		forwardButton.setFont(new java.awt.Font("Default", 0, 11));
		forwardButton.setPreferredSize(new java.awt.Dimension(100, 19));
		forwardButton.setMaximumSize(new java.awt.Dimension(100, 19));
		forwardButton.setMargin(new java.awt.Insets(2, 2, 2, 2));
		forwardButton.setMinimumSize(new java.awt.Dimension(100, 19));
		forwardButton.addActionListener(new java.awt.event.ActionListener() {

			public void actionPerformed(java.awt.event.ActionEvent evt)
			{
				forwardButtonActionPerformed(evt);
			}
		});

		navigationPanel.add(forwardButton);

		goButton.setIcon(getIconReload());
		goButton.setFont(new java.awt.Font("Default", 0, 11));
		goButton.setPreferredSize(new java.awt.Dimension(100, 19));
		goButton.setMaximumSize(new java.awt.Dimension(100, 19));
		goButton.setMargin(new java.awt.Insets(2, 2, 2, 2));
		goButton.setMinimumSize(new java.awt.Dimension(100, 19));
		goButton.addActionListener(new java.awt.event.ActionListener() {

			public void actionPerformed(java.awt.event.ActionEvent evt)
			{
				goButtonActionPerformed(evt);
			}
		});

		navigationPanel.add(goButton);

		gridBagConstraints1 = new java.awt.GridBagConstraints();
		gridBagConstraints1.gridx = 0;
		gridBagConstraints1.gridy = 0;
		gridBagConstraints1.gridwidth = 2;
		gridBagConstraints1.gridheight = 1;
		gridBagConstraints1.fill = java.awt.GridBagConstraints.HORIZONTAL;
		gridBagConstraints1.weightx = 1.0;
		add(navigationPanel, gridBagConstraints1);

		progressPanel.add(progressBar);
	}


	private ImageIcon getIconBack()
	{
		return new ImageIcon(getClass().getResource(
			"/pl/psnc/dlibra/ps/applet/viewer/res/back.gif"));
	}


	private ImageIcon getIconForward()
	{
		return new ImageIcon(getClass().getResource(
			"/pl/psnc/dlibra/ps/applet/viewer/res/forward.gif"));
	}


	private ImageIcon getIconReload()
	{
		return new ImageIcon(getClass().getResource(
			"/pl/psnc/dlibra/ps/applet/viewer/res/reload.gif"));
	}


	private String getInitialUrl(String url)
	{
		if (url.indexOf(';') == -1) {
			if (url.indexOf('?') == -1) {
				return url + ";" + SESSION_ID + "=" + sessionId;
			}
			return url.substring(0, url.indexOf('?')) + ";" + SESSION_ID + "="
					+ sessionId + url.substring(url.indexOf('?'), url.length());
		}
		return "";
	}


	private String createNewUrl(String url)
	{
		if (url.indexOf(";") < 0) {
			if (url.indexOf("Content") >= 0) {
				if (url.indexOf("#") > -1) {
					url = url.substring(0, url.indexOf("#"));
				}
				if (url.indexOf('?') < 0) {
					return url + ";" + SESSION_ID + "=" + sessionId;
				}
				return url.substring(0, url.indexOf('?')) + ";" + SESSION_ID
						+ "=" + sessionId + url.substring(url.indexOf('?'));
			}
		}
		return url;
	}


	/**
	 * Loads document from a given url.
	 * @param url
	 */
	private void loadDocument(String url)
	{
		if (url == null) {
			return;
		}

		final String reference = getReference(url);
		url = createNewUrl(url);
		statusLabel.setText(lang.getProperty("status_getting"));

		final int endIndex = url.lastIndexOf("/") > url.lastIndexOf(".") ? url
				.length() : url.lastIndexOf("/") + 1;
		ViewerHTMLEditorKit.setBaseUrl(url.substring(0, endIndex));

		final String finalURL = url;
		OperationInProgressManager.runTask(new OperationInProgress() {

			public void doOperation(OperationInProgressManager manager)
			{
				try {
					manager.start();
					String contentType = CONTENT_HTML;
					ContentBean contentBean = contentManager
							.getDocument(finalURL);

					mainEditorPane.setContentType(contentType);
					//customize editor with custom style sheet which retrieves data
					//using encrypted stream
					mainEditorPane.setDocument(new HTMLDocument(
							new CustomImportStyleSheet(utils)));

					HTMLDocument doc = (HTMLDocument) mainEditorPane
							.getDocument();
					Object content = contentBean.getContent();
					if (content instanceof String) {
						//set base for relative requests
						doc.setBase(new URL(finalURL.substring(0, endIndex)));
						mainEditorPane.setText((String) content);
					}
					else if (content instanceof URL) {
						//set base for relative requests
						doc.setBase((URL) content);
						mainEditorPane.setPage((URL) content);
					}
					else
						throw new InvalidClassException(INVALID_CLASS_ERROR);

					addressTextField.setText(finalURL);
					updateNavigation();
					statusLabel.setText(lang.getProperty("status_ready"));
				}
				catch (IOException e) {
					showErrorPage(e);
				}
				finally {
					manager.stop();
					// make sure that the upper left corner is visible
					// or selected reference is visible
					EventQueue.invokeLater(new Runnable() {

						public void run()
						{
							if (reference != null) {
								mainEditorPane.scrollToReference(reference);
							}
							else {
								mainScrollPane.getViewport()
										.scrollRectToVisible(
											new Rectangle(0, 0, 1, 1));
							}
						}
					});
				}
			}
		}, SwingUtilities.getWindowAncestor(this));
	}


	private String getReference(String url)
	{
		String reference = null;
		if (url.indexOf(";") < 0) {
			if (url.indexOf("Content") >= 0) {
				if (url.indexOf("#") > -1) {
					reference = url.substring(url.indexOf("#"));
					if (reference.length() > 1) {
						reference = reference.substring(1);
					}
					else {
						reference = null;
					}
				}
			}
		}
		return reference;
	}


	/**
	 *
	 */
	private void updateNavigation()
	{
		backButton.setEnabled(contentManager == null ? false : contentManager
				.getHistory().isBackEnabled());
		forwardButton.setEnabled(contentManager == null ? false
				: contentManager.getHistory().isForwardEnabled());

		JViewport viewport = mainScrollPane.getViewport();
		viewport.setViewPosition(new Point(0, 0));

	}
}
