/*
 * Copyright (c) 2011 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.applet.jpg.viewer;

import static pl.psnc.dlibra.applet.jpg.viewer.ViewerApplet.lang;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.image.BufferedImage;
import java.util.Formatter;

import javax.swing.AbstractButton;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JToggleButton;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import pl.psnc.dlibra.app.extension.fh.jpg.FileInfo;

/**
 * Panel that displays an image along with zoom controls, text area for
 * displaying description or comment and a status bar for showing messages
 * and errors.
 */
public class ImageViewPanel
	extends JPanel
{
	
	final static double MOUSE_WHEEL_ZOOM_STEP = 1.1;

	final static int POSITION_LIMIT_SPACE = 30;

	final static int ZOOM_MIN_SIZE = 30;

	BufferedImage image;

	FileInfo fileInfo;

	ZoomSlider zoomSlider;

	/**
	 * Coordinates of the center of the view. The origin is in the top left
	 * corner of the image, the unit is pixel on the screen.
	 */
	int x, y;

	private MouseAdapter mouseAdapter = new MouseAdapter() {

		private int mode = 0; //0-none, 1-drag

		private int prevX, prevY;


		@Override
		public void mousePressed(MouseEvent e)
		{
			if (mode == 0 && e.getButton() == MouseEvent.BUTTON1) {
				mode = 1;
				prevX = e.getX();
				prevY = e.getY();
			}
		}


		@Override
		public void mouseReleased(MouseEvent e)
		{
			if (mode == 1 && e.getButton() == MouseEvent.BUTTON1) {
				mode = 0;
			}
		}


		@Override
		public void mouseDragged(MouseEvent e)
		{
			if (mode == 1) {
				x -= e.getX() - prevX;
				prevX = e.getX();
				y -= e.getY() - prevY;
				prevY = e.getY();
				limitPositionAndZoom();
				repaint();
			}
		}


		public void mouseWheelMoved(MouseWheelEvent e)
		{
			int clicks = e.getWheelRotation();
			double zoomChange = 1;
			for (; clicks > 0; clicks--) {
				zoomChange /= MOUSE_WHEEL_ZOOM_STEP;
			}
			for (; clicks < 0; clicks++) {
				zoomChange *= MOUSE_WHEEL_ZOOM_STEP;
			}

			zoomSlider.changeZoom(zoomChange);
		};
	};

	private JPanel infoPanel;

	private JLabel infoHeader;

	private JTextArea infoArea;

	private JLabel statusLabel;

	int infoContent = 0;//0-nothing, 1-description, 2-comment

	private boolean dimmed;


	public ImageViewPanel()
	{
		addMouseListener(mouseAdapter);
		addMouseMotionListener(mouseAdapter);
		addMouseWheelListener(mouseAdapter);

		addComponentListener(new ComponentAdapter() {
			@Override
			public void componentResized(ComponentEvent e)
			{
				limitPositionAndZoom();
			}
		});

		setLayout(new BorderLayout());
		add(createLeftPanel(), BorderLayout.WEST);
		add(createRightPanel(), BorderLayout.EAST);

		statusLabel = new JLabel();
		statusLabel.setOpaque(true);
		add(statusLabel, BorderLayout.SOUTH);
	}


	/**
	 * @return
	 */
	private JPanel createLeftPanel()
	{
		JPanel leftPanel = new JPanel();
		leftPanel.setOpaque(false);
		leftPanel.setLayout(new BoxLayout(leftPanel, BoxLayout.PAGE_AXIS));
	
		JToggleButton zoomButton = new JToggleButton(new ImageIcon(getClass()
				.getResource("/pl/psnc/dlibra/applet/jpg/viewer/res/zoom.png")));
		zoomButton.setFocusPainted(false);
		zoomButton.setAlignmentX(0.5f);
		zoomButton.setToolTipText(lang.getProperty("zoomButton.tooltip"));
		leftPanel.add(zoomButton);

		final JPanel zoomPanel = new JPanel();
		zoomPanel.setAlignmentX(0.5f);
		zoomPanel.setLayout(new BoxLayout(zoomPanel, BoxLayout.PAGE_AXIS));
		zoomPanel.setBorder(BorderFactory.createLineBorder(Color.black));
		zoomPanel.setVisible(false);

		final JLabel zoomLabel = new JLabel("100%");
		zoomLabel.setAlignmentX(0.5f);
		zoomPanel.add(zoomLabel);

		zoomSlider = new ZoomSlider();
		zoomSlider.setMaximumSize(zoomSlider.getPreferredSize());
		zoomSlider.setAlignmentX(0.5f);
		zoomPanel.add(zoomSlider);

		leftPanel.add(zoomPanel);
		leftPanel.add(Box.createVerticalGlue());

		zoomButton.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e)
			{
				zoomPanel.setVisible(((AbstractButton) e.getSource())
						.isSelected());
			}
		});
		zoomSlider.addChangeListener(new ChangeListener() {
			
			double previousZoom = zoomSlider.getZoom();
	
	
			@Override
			public void stateChanged(ChangeEvent e)
			{
				double newZoom = zoomSlider.getZoom();
				double scale = newZoom / previousZoom;
				previousZoom = newZoom;
				x *= scale;
				y *= scale;
				zoomLabel.setText(new Formatter().format("%.2f%%",
					newZoom * 100).toString());
				limitPositionAndZoom();
				repaint();
			}
		});
		return leftPanel;
	}


	private JPanel createRightPanel()
	{
		JPanel rightPanel = new JPanel();
		rightPanel.setLayout(new BoxLayout(rightPanel, BoxLayout.PAGE_AXIS));
		rightPanel.setOpaque(false);
		rightPanel.setPreferredSize(new Dimension(300, 300));

		JPanel buttonsPanel = new JPanel();
		buttonsPanel.setLayout(new BoxLayout(buttonsPanel, BoxLayout.LINE_AXIS));
		buttonsPanel.setAlignmentX(1.0f);

		final JToggleButton descriptionButton = new JToggleButton(new ImageIcon(getClass()
			.getResource("/pl/psnc/dlibra/applet/jpg/viewer/res/description.png")));
		descriptionButton.setToolTipText(lang.getProperty("descriptionButton.tooltip"));
		buttonsPanel.add(descriptionButton);

		final JToggleButton commentButton = new JToggleButton(new ImageIcon(getClass()
			.getResource("/pl/psnc/dlibra/applet/jpg/viewer/res/comment.png")));
		commentButton.setToolTipText(lang.getProperty("commentButton.tooltip"));
		buttonsPanel.add(commentButton);
		rightPanel.add(buttonsPanel);

		infoPanel = new JPanel();
		infoPanel.setBackground(Color.white);
		infoPanel.setBorder(BorderFactory.createLineBorder(Color.black));
		infoPanel.setLayout(new BorderLayout());
		infoPanel.setVisible(false);
		infoPanel.setMaximumSize(new Dimension(300, 300));
		rightPanel.add(infoPanel);
		infoPanel.setAlignmentX(1.0f);

		infoHeader = new JLabel();
		infoPanel.add(infoHeader, BorderLayout.NORTH);

		infoArea = new JTextArea();
		infoArea.setLineWrap(true);
		infoArea.setWrapStyleWord(true);
		infoArea.setEnabled(false);
		infoArea.setDisabledTextColor(infoArea.getForeground());
		JScrollPane scrollPane = new JScrollPane();
		scrollPane.setHorizontalScrollBarPolicy(
			ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		scrollPane.setBorder(BorderFactory.createEmptyBorder());
		scrollPane.setViewportView(infoArea);
		infoPanel.add(scrollPane, BorderLayout.CENTER);

		ActionListener buttonListener = new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e)
			{
				JToggleButton button = (JToggleButton) e.getSource();
				if (!button.isSelected()) {
					infoContent = 0;
				}
				else {
					boolean description = button == descriptionButton;
					JToggleButton otherButton = description ? commentButton : descriptionButton;
					otherButton.setSelected(false);
					infoContent = description ? 1 : 2;
				}
				refreshInfoPanel();
			}
		};
		descriptionButton.addActionListener(buttonListener);
		commentButton.addActionListener(buttonListener);

		rightPanel.add(Box.createGlue());
		return rightPanel;
	}


	void refreshInfoPanel()
	{
		if (infoContent == 0) {
			infoPanel.setVisible(false);
		}
		else {
			infoPanel.setVisible(true);
			infoHeader.setText(ViewerApplet.lang
					.getProperty(infoContent == 1 ? "info.header.description"
							: "info.header.comment"));
			infoArea.setText(infoContent == 1 ? fileInfo.getDescription()
					: fileInfo.getComment());
		}
		SwingUtilities.invokeLater(new Runnable() {
			
			@Override
			public void run()
			{
				revalidate();
			}
		});
	}


	@Override
	protected void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		if (image == null) {
			return;
		}
		Graphics2D g2 = (Graphics2D) g;
		g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
			RenderingHints.VALUE_INTERPOLATION_BILINEAR);
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			RenderingHints.VALUE_ANTIALIAS_ON);
		final double zoom = zoomSlider.getZoom();
		g2.drawImage(image, 0, 0, getWidth(), getHeight(),
			(int) ((x - getWidth() / 2) / zoom),
			(int) ((y - getHeight() / 2) / zoom),
			(int) ((x + getWidth() / 2) / zoom),
			(int) ((y + getHeight() / 2) / zoom), null);

		if (dimmed) {
			g2.setColor(new Color(0, 0, 0, 0.5f));
			g2.fillRect(0, 0, getWidth(), getHeight());
		}
	}


	public BufferedImage getImage()
	{
		return image;
	}


	public void setImage(BufferedImage image, FileInfo fileInfo)
	{
		this.image = image;
		this.fileInfo = fileInfo;
		zoomSlider.resetZoom(image.getWidth(), image.getHeight(),
			getWidth(), getHeight());
		this.x = (int) (image.getWidth() / 2 * zoomSlider.getZoom());
		this.y = (int) (image.getHeight() / 2 * zoomSlider.getZoom());
		refreshInfoPanel();
		repaint();
	}


	void limitPositionAndZoom() {
		if (image == null) {
			return;
		}
		final double zoom = zoomSlider.getZoom();
		if (image.getWidth() * zoom < ZOOM_MIN_SIZE) {
			zoomSlider.setZoom(1.0 * ZOOM_MIN_SIZE / image.getWidth());
		}
		if (image.getHeight() * zoom < ZOOM_MIN_SIZE) {
			zoomSlider.setZoom(1.0 * ZOOM_MIN_SIZE / image.getHeight());
		}
		if (image.getWidth() * zoom + 2 * POSITION_LIMIT_SPACE < getWidth()) {
			x = (int) (image.getWidth() / 2 * zoom);
		}
		else {
			int leftSpace = getWidth() / 2 - x;
			if (leftSpace > POSITION_LIMIT_SPACE) {
				x = getWidth() / 2 - POSITION_LIMIT_SPACE;
			}
			int rightSpace = (int) (getWidth() / 2 - image.getWidth() * zoom + x);
			if (rightSpace > POSITION_LIMIT_SPACE) {
				x = (int) (POSITION_LIMIT_SPACE - getWidth() / 2 + image
						.getWidth() * zoom);
			}
		}
		if (image.getHeight() * zoom + 2 * POSITION_LIMIT_SPACE < getHeight()) {
			y = (int) (image.getHeight() / 2 * zoom);
		}
		else {
			int topSpace = getHeight() / 2 - y;
			if (topSpace > POSITION_LIMIT_SPACE) {
				y = getHeight() / 2 - POSITION_LIMIT_SPACE;
			}
			int bottomSpace = (int) (getHeight() / 2 - image.getHeight() * zoom + y);
			if (bottomSpace > POSITION_LIMIT_SPACE) {
				y = (int) (POSITION_LIMIT_SPACE - getHeight() / 2 + image
						.getHeight() * zoom);
			}
		}
	}


	public void showMessage(String message)
	{
		statusLabel.setText(message);
	}


	public void setDimmed(boolean dimmed)
	{
		this.dimmed = dimmed;
	}
}