package pl.psnc.dlibra.applet.jpg.viewer;

import static pl.psnc.dlibra.applet.jpg.viewer.ViewerApplet.lang;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;

import javax.imageio.ImageIO;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.SwingWorker;

import pl.psnc.dlibra.app.extension.fh.jpg.FileInfo;
import pl.psnc.util.crypto.CryptoUtils;


/**
 * Main panel of the applet, conatins areas for thumbnails and
 * for selected image. 
 */
public class JPGViewerPanel
	extends JPanel
{

	private JSplitPane splitPane = new JSplitPane();

	private ImageViewPanel imageViewPanel = new ImageViewPanel();

	private ThumbnailsViewPanel thumbnailsViewPanel = new ThumbnailsViewPanel();

	private List<FileInfo> fileInfos;

	private SwingWorker<?, ?> downloadWorker;

	String initialUrl;

	String sessionId;

	CryptoUtils utils;

	ConnectionManager connectionManager;

	Watermark watermark = null;
	

	private void initComponents()
	{
		thumbnailsViewPanel.addSelectionListener(new ActionListener() {

			FileInfo previous = null;


			@Override
			public void actionPerformed(ActionEvent e)
			{
				FileInfo fileInfo = (FileInfo) e.getSource();
				if (fileInfo.equals(previous)) {
					return;
				}
				loadImage(fileInfo);
				previous = fileInfo;
			}
		});

		this.setLayout(new BorderLayout());
		this.add(splitPane, BorderLayout.CENTER);
		splitPane.setLeftComponent(thumbnailsViewPanel);
		splitPane.setRightComponent(imageViewPanel);
	}


	public JPGViewerPanel()
	{
		initComponents();
	}


	public void setConnectionManager(ConnectionManager connectionManager)
		throws IOException
	{
		this.connectionManager = connectionManager;
		try {
			fileInfos = connectionManager.downloadFileInfos();
			if (connectionManager.isSecured()) {
				Properties watermarkProperties = connectionManager
						.downloadWatermarkProperties();
				watermark = new Watermark();
				watermark.init(watermarkProperties, connectionManager);
			}
		}
		catch (IOException e) {
			fileInfos = Collections.emptyList();
			throw e;
		}
	}


	public void start()
	{
		thumbnailsViewPanel.downloadThumbnails(fileInfos, connectionManager);
	}


	/**
	 * @param fileInfo
	 */
	void loadImage(final FileInfo fileInfo)
	{
		showMessage(lang.getProperty("loading.message")
				+ fileInfo.getFilePath());
		imageViewPanel.setDimmed(true);
		if (downloadWorker != null && !downloadWorker.isDone()) {
			downloadWorker.cancel(true);
		}
		downloadWorker = new SwingWorker<BufferedImage, BufferedImage>() {

			@Override
			protected BufferedImage doInBackground()
				throws Exception
			{
				InputStream is = connectionManager.getImageStream(fileInfo);
				BufferedImage image = ImageIO.read(is);
				is.close();

				if (isCancelled()) {
					return null;
				}

				if (watermark != null) {
					image = watermark.apply(image);
				}
				return image;
			}


			@Override
			protected void done()
			{
				try {
					imageViewPanel.setImage(get(), fileInfo);
					showMessage("");
					imageViewPanel.setDimmed(false);
				}
				catch (CancellationException e) {
					// nothing to do
				}
				catch (ExecutionException e) {
					showError(e);
				}
				catch (Exception e) {
					throw new RuntimeException("This shouldn't happen", e);
				}
			}
		};
		downloadWorker.execute();
	}


	public void showError(Throwable e)
	{
		imageViewPanel.showMessage(lang.getProperty("error.message")
				+ e.getMessage());
		e.printStackTrace();
	}


	public void showMessage(String message)
	{
		imageViewPanel.showMessage(message);
	}
}
