package pl.psnc.dlibra.applet.jpg.viewer;

import static pl.psnc.dlibra.applet.jpg.viewer.ViewerApplet.lang;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import javax.imageio.ImageIO;
import javax.swing.AbstractButton;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JToggleButton;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;

import pl.psnc.dlibra.app.extension.fh.jpg.FileInfo;

/**
 * Panel displaying thumbnails of all images in digital publication.
 */
public class ThumbnailsViewPanel
	extends JPanel
{

	volatile int thumbnailWidth = 300;

	volatile int thumbnailHeight = 300;

	JScrollPane scrollPane;

	List<ActionListener> selectionListeners = new ArrayList<ActionListener>();

	List<FileInfo> fileInfos;

	List<BufferedImage> images = new Vector<BufferedImage>();

	List<AbstractButton> thumbnails = new ArrayList<AbstractButton>();

	private ImageIcon loadingIcon = new ImageIcon(getClass().getResource(
			"/pl/psnc/dlibra/applet/jpg/viewer/res/loading.gif"));

	JPanel thumbnailsPanel = new JPanel();

	JButton selectPrevButton;

	JButton selectNextButton;

	JComboBox selectionCombo;

	private ActionListener thumbnailListener = new ActionListener() {

		@Override
		public void actionPerformed(ActionEvent e)
		{
			AbstractButton thumbnailButton = (AbstractButton) e.getSource();
			if (!thumbnailButton.isSelected()) {
				// the clicked button had been selected, which means the selection hasn't changed
				thumbnailButton.setSelected(true);
				return;
			}
			int index = thumbnails.indexOf(thumbnailButton);
			changeSelection(index);
		}
	};

	public ThumbnailsViewPanel()
	{
		thumbnailsPanel = new JPanel();
		thumbnailsPanel.setBorder(BorderFactory.createEmptyBorder());
		thumbnailsPanel.setLayout(new GridBagLayout());
		JPanel middlePanel = new JPanel();
		middlePanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
		middlePanel.setBorder(BorderFactory.createEmptyBorder());
		middlePanel.add(thumbnailsPanel);
		scrollPane = new JScrollPane(middlePanel,
				JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
				JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		scrollPane.setBorder(BorderFactory.createEmptyBorder());
		scrollPane.addComponentListener(new ComponentAdapter() {

			@Override
			public void componentResized(ComponentEvent e)
			{
				if (fileInfos == null) {
					return;
				}
				for (int i = 0; i < fileInfos.size(); i++) {
					recreateIcon(i);
				}
				thumbnailsPanel.invalidate();
			}
		});
		scrollPane.getVerticalScrollBar().addAdjustmentListener(
			new AdjustmentListener() {

				@Override
				public void adjustmentValueChanged(AdjustmentEvent e)
				{
					synchronized (scrollPane) {
						scrollPane.notify();
					}
				}
			});
		this.setBorder(BorderFactory.createEmptyBorder());
		this.setLayout(new BorderLayout());
		this.add(scrollPane, BorderLayout.CENTER);
		this.setPreferredSize(new Dimension(200, 300));

		JPanel topPanel = new JPanel();
		topPanel.setLayout(new BorderLayout());
		selectPrevButton = new JButton(new ImageIcon(getClass()
			.getResource("/pl/psnc/dlibra/applet/jpg/viewer/res/previous.png")));
		selectPrevButton.setToolTipText(lang.getProperty("previousButton.tooltip"));
		selectPrevButton.setMargin(new Insets(1, 1, 1, 1));
		selectPrevButton.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e)
			{
				int current = -1;
				for (int i = 0; i < thumbnails.size(); i++) {
					if (thumbnails.get(i).isSelected()) {
						current = i;
						break;
					}
				}
				if (current > 0) {
					changeSelection(current - 1);
				}
			}
		});
		topPanel.add(selectPrevButton, BorderLayout.WEST);
		selectNextButton = new JButton(new ImageIcon(getClass()
			.getResource("/pl/psnc/dlibra/applet/jpg/viewer/res/next.png")));
		selectNextButton.setToolTipText(lang.getProperty("nextButton.tooltip"));
		selectNextButton.setMargin(new Insets(1, 1, 1, 1));
		selectNextButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e)
			{
				int current = -1;
				for (int i = 0; i < thumbnails.size(); i++) {
					if (thumbnails.get(i).isSelected()) {
						current = i;
						break;
					}
				}
				if (current >= 0) {
					changeSelection(current + 1);
				}
			}
		});
		topPanel.add(selectNextButton, BorderLayout.EAST);
		selectionCombo = new JComboBox();
		selectionCombo.setMinimumSize(selectionCombo.getMinimumSize());
		selectionCombo.addItemListener(new ItemListener() {

			@Override
			public void itemStateChanged(ItemEvent e)
			{
				if (e.getStateChange() != ItemEvent.SELECTED) {
					return;
				}
				Object selectedItem = e.getItem();
				for (int i = 0; i < selectionCombo.getItemCount(); i++) {
					if (selectionCombo.getItemAt(i).equals(selectedItem)) {
						changeSelection(i);
						break;
					}
				}
			}
		});
		topPanel.add(selectionCombo, BorderLayout.CENTER);
		this.add(topPanel, BorderLayout.NORTH);
	}


	/**
	 * Adds listener that will be notified when user clicks on a thumbnail.
	 * @param selectionListener action listener. When invoked, event source will
	 * be {@link FileInfo} of the file selected.
	 */
	public void addSelectionListener(ActionListener selectionListener)
	{
		selectionListeners.add(selectionListener);
	}


	/**
	 * @see ThumbnailsViewPanel#addSelectionListener(ActionListener)
	 */
	public boolean removeSelectionListener(ActionListener selectionListener)
	{
		return selectionListeners.remove(selectionListener);
	}


	public void downloadThumbnails(final List<FileInfo> fileInfos,
			final ConnectionManager connectionManager)
	{
		this.fileInfos = fileInfos;
		thumbnailsPanel.removeAll();
		thumbnails.clear();
		images.clear();
		images.addAll(Arrays.asList(new BufferedImage[fileInfos.size()]));

		createThumbnailButtons(fileInfos);

		selectionCombo.removeAllItems();
		for (int i = 0; i < fileInfos.size(); i++) {
			selectionCombo.addItem("" + (i + 1) + " - "
					+ fileInfos.get(i).getFilePath());
		}
		selectPrevButton.setEnabled(false);
		selectNextButton.setEnabled(fileInfos.size() > 1);

		new SwingWorker<Object, Integer>() {

			private final static int THUMBNAILS_BUFFER_SIZE = 30;

			// items in backward direction are considered to lay DIRECTION_BIAS times further
			private final static int DIRECTION_BIAS = 2;

			private Hashtable<Integer, String> errors = new Hashtable<Integer, String>();

			private ImageIcon errorIcon;

			// keep track of downloaded thumbnails separatly from ThumbnailsViewPanel#images
			// because we don't care if download fails (we shouldn't redownload anyway)
			private boolean downloaded[];

			private int downloadedCount = 0;

			@Override
			protected BufferedImage doInBackground()
				throws Exception
			{
				downloaded = new boolean[fileInfos.size()];
				while (true) {
					int toDownload = getFileToDownloadWaiting();
					if (toDownload < 0) {
						// everything downloaded!
						return null;
					}
					FileInfo fileInfo = fileInfos.get(toDownload);
					InputStream is = null;
					try {
						is = connectionManager.getThumbnailStream(fileInfo);
						try {
							BufferedImage image = ImageIO.read(is);
							if (image == null) {
								throw new IOException(
										lang.getProperty("download.error.unrecognized"));
							}
							thumbnailWidth = image.getWidth();
							thumbnailHeight = image.getHeight();
							images.set(toDownload, image);
						}
						finally {
							is.close();
						}
					}
					catch (IOException e) {
						errors.put(toDownload, e.getMessage());
						e.printStackTrace();
					}
					publish(toDownload);
				}
			}


			private int getFileToDownloadWaiting()
			{
				if (downloadedCount >= fileInfos.size()) {
					return -1;
				}
				while (true) {
					int result = getFileToDownload();
					if (result >= 0) {
						return result;
					}
					synchronized (scrollPane) {
						try {
							// will be notified when scroll position changes
							scrollPane.wait();
						}
						catch (InterruptedException e) {
							return -1;
						}
					}
				}
			}


			private int getFileToDownload()
			{
				int current = getFirstVisibleThumbnail();
				Integer toRemove = null;
				if (downloadedCount >= THUMBNAILS_BUFFER_SIZE) {
					// remove item that is most far away from current position
					int maxDist = 0;
					for (int i = 0; i < fileInfos.size(); i++) {
						if (downloaded[i]) {
							int dist = i < current ? (current - i)
									* DIRECTION_BIAS : (i - current);
							if (dist > maxDist) {
								maxDist = dist;
								toRemove = i;
							}
						}
					}
					if (toRemove != null) {
						downloaded[toRemove] = false;
					}
				}

				Integer toDownload = null;
				// download item that is the closest to current position
				toDownloadLoop:
				for (int i = 0; i < fileInfos.size(); i++) {
					for (int j = 0; j < DIRECTION_BIAS; j++) {
						int index = current + i * DIRECTION_BIAS + j;
						if (index < fileInfos.size() && !downloaded[index]) {
							toDownload = index;
							break toDownloadLoop;
						}
					}
					int index = current - i;
					if (index >= 0 && !downloaded[index]) {
						toDownload = index;
						break toDownloadLoop;
					}
				}

				if (toDownload.equals(toRemove)) {
					downloaded[toRemove] = true;
					return -1;
				}
				else if (toRemove != null) {
					downloadedCount--;
					removeImage(toRemove);
				}

				downloaded[toDownload] = true;
				downloadedCount++;
				return toDownload;
			}


			private int getFirstVisibleThumbnail()
			{
				try {
					final int[] result = new int[1];
					SwingUtilities.invokeAndWait(new Runnable() {

						@Override
						public void run()
						{
							int scrollPosition = scrollPane.getViewport()
									.getViewPosition().y;
							for (int i = 1; i < thumbnails.size(); i++) {
								if (thumbnails.get(i).getY() > scrollPosition) {
									result[0] = i - 1;
									break;
								}
							}
						}
					});
					return result[0];
				}
				catch (Exception e) {
					throw new RuntimeException("This should never happen", e);
				}
			}


			protected void process(List<Integer> chunks)
			{
				for (Integer index : chunks) {
					if (images.get(index) == null) {
						if (errorIcon == null) {
							try {
								errorIcon = new ImageIcon(
										ImageIO.read(getClass().getResource(
												"/pl/psnc/dlibra/applet/jpg/viewer/res/error.png")));
							}
							catch (IOException e) {
								throw new RuntimeException(
										"this shouldn't happen");
							}
						}
						thumbnails.get(index).setIcon(errorIcon);
						thumbnails.get(index).setToolTipText(
							lang.getProperty("download.error.tooltip")
									+ errors.get(index));
					}
					recreateIcon(index);
				}
			}
		}.execute();
	}


	void removeImage(Integer index)
	{
		images.set(index, null);
		thumbnails.get(index).setIcon(loadingIcon);
		recreateIcon(index);
	}


	private void createThumbnailButtons(final List<FileInfo> fileInfos)
	{
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.fill = GridBagConstraints.NONE;
		gbc.gridx = 0;
		gbc.insets = new Insets(3, 4, 3, 4);
		for (int i = 0; i < fileInfos.size(); i++) {
			AbstractButton button = new JToggleButton();
			button.setBackground(Color.white);
			button.setForeground(Color.black);
			button.setMargin(new Insets(1, 1, 1, 1));
			button.setIcon(loadingIcon);
			button.setToolTipText(fileInfos.get(i).getFilePath());
			if (thumbnails.isEmpty()) {
				button.setSelected(true);
			}
			button.addActionListener(thumbnailListener);
			thumbnailsPanel.add(button, gbc);
			thumbnails.add(button);
			recreateIcon(i);
		}
	}


	void changeSelection(int index)
	{
		for (int i = 0; i < thumbnails.size(); i++) {
			thumbnails.get(i).setSelected(i == index);
		}
		if (index >= 0) {
			ActionEvent event = new ActionEvent(fileInfos.get(index), 0,
					null);
			for (ActionListener listener : selectionListeners) {
				listener.actionPerformed(event);
			}
		}
		selectionCombo.setSelectedIndex(index);
		selectPrevButton.setEnabled(index > 0);
		selectNextButton.setEnabled(index < thumbnails.size() - 1);

		if (thumbnails.get(index).getVisibleRect().isEmpty()) {
			scrollPane.getViewport().setViewPosition(
				thumbnails.get(index).getLocation());
		}
	}


	void recreateIcon(int index)
	{
		int maxWidth = scrollPane.getViewportBorderBounds().width;
		AbstractButton thumbnail = thumbnails.get(index);
		int insets = 8; //padding from gridbaglayout in thumbnailsPanel
		if (images.get(index) == null) {
			// thumbnail has not been loaded yet, there's a loading icon
			int size = Math.min(maxWidth, thumbnailWidth) - insets;
			thumbnail.setPreferredSize(new Dimension(size, size));
			return;
		}
		else {
			thumbnail.setPreferredSize(null);
		}

		BufferedImage image = images.get(index);
		BufferedImage scaledImage = image;

		Insets borderInsets = thumbnail.getBorder().getBorderInsets(thumbnail);
		insets += borderInsets.left + borderInsets.right;
		maxWidth = Math.max(maxWidth - insets, 20);

		if (maxWidth < image.getWidth()) {
			int maxHeight = image.getHeight() * maxWidth / image.getWidth();
			scaledImage = getScaledInstance(image, maxWidth, maxHeight);
		}
		ImageIcon icon = new ImageIcon(scaledImage);
		thumbnail.setIcon(icon);

		scrollPane.getVerticalScrollBar().setUnitIncrement(
			thumbnail.getHeight() + 6); //six is from gridbaglayout in thumbnailsPanel
	}


	/**
	 * Resizes given image, assuring good quality when target size is much smaller than
	 * source size. Simplified version of code from
	 * http://today.java.net/pub/a/today/2007/04/03/perils-of-image-getscaledinstance.html
	 */
	public static BufferedImage getScaledInstance(BufferedImage img,
			int targetWidth, int targetHeight)
	{
		int type = BufferedImage.TYPE_INT_RGB;
		BufferedImage ret = img;
		int w = img.getWidth();
		int h = img.getHeight();

		do {
			if (w > targetWidth) {
				w /= 2;
			}
			if (w < targetWidth) {
				w = targetWidth;
			}

			if (h > targetHeight) {
				h /= 2;
			}
			if (h < targetHeight) {
				h = targetHeight;
			}

			BufferedImage tmp = new BufferedImage(w, h, type);
			Graphics2D g2 = tmp.createGraphics();
			g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
				RenderingHints.VALUE_INTERPOLATION_BILINEAR);
			g2.drawImage(ret, 0, 0, w, h, null);
			g2.dispose();

			ret = tmp;
		}
		while (w != targetWidth || h != targetHeight);

		return ret;
	}

}
