/*
 * Copyright (c) 2011 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.applet.jpg.viewer;

import java.util.Dictionary;
import java.util.Hashtable;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JSlider;

/**
 * Slider modified to support image zoom.
 */
public class ZoomSlider
	extends JSlider
{
	private static final int MAX_VALUE = 200;

	private static final double MIN_ZOOM = 1.0 / 8;

	private static final double MAX_ZOOM = 8.0;

	private double zoom = 1.0;

	boolean changeInProgress = false;


	public ZoomSlider()
	{
		super(JSlider.VERTICAL);
		setMaximum(MAX_VALUE);
	}


	public double getZoom()
	{
		if (getValue() != zoomToValue(zoom)) {
			// slider position changed, but the new zoom has not yet been stored in the zoom field
			return valueToZoom(getValue());
		}
		return zoom;
	}


	public void setZoom(double zoom)
	{
		this.zoom = zoom;
		if (!changeInProgress) {
			changeInProgress = true;
			setValue(zoomToValue(zoom));
			if (getValue() == getMinimum() || getValue() == getMaximum()) {
				fireStateChanged();
			}
			changeInProgress = false;
		}
	}


	public void changeZoom(double multiplier)
	{
		setZoom(zoom * multiplier);
	}


	public void resetZoom(int objectWidth, int objectHeight,
			int containerWidth, int containerHeight)
	{
		Dictionary<Integer, JComponent> labels = new Hashtable<Integer, JComponent>();
		labels.put(zoomToValue(1.0/8), new JLabel("x0.125"));
		labels.put(zoomToValue(1.0/4), new JLabel("x0.25"));
		labels.put(zoomToValue(1.0/2), new JLabel("x0.5"));
		labels.put(zoomToValue(1.0), new JLabel("x1"));
		labels.put(zoomToValue(2.0), new JLabel("x2"));
		labels.put(zoomToValue(4.0), new JLabel("x4"));
		labels.put(zoomToValue(8.0), new JLabel("x8"));
		setLabelTable(labels);
		setPaintLabels(true);
		setMaximumSize(getPreferredSize());

		double startingZoom = Math.min(1.0 * containerWidth / objectWidth,
			1.0 * containerHeight / objectHeight);
		setZoom(startingZoom);
	}


	@Override
	protected void fireStateChanged()
	{
		if (!changeInProgress) {
			changeInProgress = true;
			setZoom(valueToZoom(getValue()));
			changeInProgress = false;
		}
		super.fireStateChanged();
	}

	private int zoomToValue(double zoom)
	{
		double log = Math.log(zoom) / Math.log(2);
		double maxLog = Math.log(MAX_ZOOM) / Math.log(2);
		double minLog = Math.log(MIN_ZOOM) / Math.log(2);
		log = Math.max(Math.min(log, maxLog), minLog);
		double value = MAX_VALUE * (log - minLog) / (maxLog - minLog);
		return (int) value;
	}


	private double valueToZoom(int value)
	{
		double maxLog = Math.log(MAX_ZOOM) / Math.log(2);
		double minLog = Math.log(MIN_ZOOM) / Math.log(2);
		double log = minLog + 1.0 * value / MAX_VALUE * (maxLog - minLog);
		double zoom = Math.pow(2, log);
		return zoom;
	}
}
